from .nonce import NewNonceMessage
from .account import NewAccountMessage
from .order import NewOrderMessage, OrderFinalizationMessage, OrderMessage
from .authz import AuthorizationMessage
from .challenge import ChallengeMessage
from .certificate import CertMessage



import re

import inject
import josepy
import josepy.jwk

from acmev2.errors import ACMEError
from acmev2.messages import NewAccountMessage
from acmev2.models import AccountResource
from acmev2.services import (
    ACMEEndpoint,
    IAccountService,
    IDirectoryService,
)
from acmev2.settings import ACMESettings

from .base import ACMEModelResponse, ACMERequestHandler

EMAIL_REGEX = re.compile("^.+@.+$")
# Caddy sends mailto:default as a email contact token
# We'll just remove that.
EMAIL_DEFAULT = "mailto:default"


class NewAccountRequestHandler(ACMERequestHandler):
    """Creates a new account or retrieves an account id if the passed in jwk matches an existing one."""

    message_type = NewAccountMessage
    settings = inject.attr(ACMESettings)

    @inject.autoparams()
    def process(
        self,
        msg: NewAccountMessage,
        directory_service: IDirectoryService,
        account_service: IAccountService,
    ):
        self.validate_embedded_jwk(msg)
        self.validate_contact(msg)

        if existing_account := account_service.get_by_jwk(jwk=msg.jwk):
            return ACMEModelResponse(
                msg=existing_account,
                code=200,
                location=directory_service.url_for(
                    ACMEEndpoint.account, existing_account.id
                ),
            )

        if msg.payload.onlyReturnExisting:
            raise ACMEError(type="accountDoesNotExist")

        self.validate_tos(msg)
        eab_kid: str = None
        if msg.payload.externalAccountBinding:
            if not self.verify_eab(msg):
                raise ACMEError(type="malformed")

            eab_kid = msg.payload.externalAccountBinding.protected_decoded.get("kid")
        elif self.settings.eab_required:
            raise ACMEError(type="externalAccountRequired")

        resource = AccountResource(
            status="valid",
            termsOfServiceAgreed=msg.payload.termsOfServiceAgreed,
            contact=msg.payload.contact,
            jwk=msg.jwk,
        )

        resource = account_service.create(resource, eab_kid)

        return ACMEModelResponse(
            msg=resource,
            code=201,
            location=directory_service.url_for(ACMEEndpoint.account, resource.id),
        )

    @inject.autoparams()
    def verify_eab(
        self, msg: NewAccountMessage, account_service: IAccountService
    ) -> bool:
        eab = msg.payload.externalAccountBinding
        # The eab contains a jwk
        eab_jwk: josepy.jwk.JWK = josepy.JWK.from_json(
            msg.payload.externalAccountBinding.payload_decoded
        )
        # The eab jwk must be the same as the outer jwk
        if msg.jwk.thumbprint() != eab_jwk.thumbprint():
            raise ACMEError()

        jws = josepy.JWS.from_json(eab.model_dump())
        # The hmac key is pulled from the account using the passed in kid
        hmac_key = account_service.get_eab_hmac(eab.protected_decoded.get("kid"))
        if not hmac_key:
            raise ACMEError()

        # hmac key is used to verify the eab jws
        jwk = josepy.JWK.from_json({"k": hmac_key, "kty": "oct"})

        return jws.verify(jwk)

    def validate_embedded_jwk(self, msg: NewAccountMessage):
        if not msg.has_embedded_jwk:
            raise ACMEError(detail="Embedded JWK expected.")

    def validate_contact(self, msg: NewAccountMessage):
        # Only mailto contacts are currently supported
        mailtos = [c for c in msg.payload.contact or [] if c.startswith("mailto:")]

        # Remove 'default' mailtos
        mailtos = [c for c in mailtos if c != EMAIL_DEFAULT]

        for email in mailtos:
            email = email.lstrip("mailto:").strip()

            if EMAIL_REGEX.match(email):
                continue

            raise ACMEError(
                type="unsupportedContact",
                detail=f"Contact {email} is unsupported.",
            )

    @inject.autoparams()
    def validate_tos(
        self,
        msg: NewAccountMessage,
        directory_service: IDirectoryService,
    ):
        payload = msg.payload
        if not payload.termsOfServiceAgreed:
            directory = directory_service.get_directory()
            raise ACMEError(
                type="userActionRequired",
                detail="User must agree to terms of service.",
                header_links=[
                    (
                        directory["meta"]["termsOfService"],
                        "terms-of-service",
                    )
                ],
            )



from acmev2.models import NewOrderRequestSchema, OrderFinalizationRequestSchema
from acmev2.models import CertResource
from acmev2.services import ACMEResourceType, ACMEEndpoint
from .base import SignedACMEMessage, AuthenticatedACMEMessage, WithResourceUrl


class CertMessage(
    SignedACMEMessage[NewOrderRequestSchema],
    WithResourceUrl[CertResource],
):
    """https://datatracker.ietf.org/doc/html/rfc8555/#section-7.4.2"""

    protected_resource_type = ACMEResourceType.cert
    protected_url_type = ACMEEndpoint.cert



import random
import josepy
import re
import string
from datetime import datetime

import inject
import josepy
from cryptography import x509

from acmev2.errors import ACMEBadCSRDetail, ACMEBadCSRError, ACMEError
from acmev2.messages import NewOrderMessage, OrderFinalizationMessage, OrderMessage
from acmev2.models import (
    AuthorizationResource,
    AuthorizationStatus,
    HTTPChallengeResource,
    CustomChallengeResource,
    IdentifierType,
    OrderResource,
    OrderStatus,
    without,
)
from acmev2.services import (
    ACMEEndpoint,
    IAuthorizationService,
    IChallengeService,
    IDirectoryService,
    IOrderService,
)
from acmev2.settings import ACMESettings, Challenges

from .base import ACMEModelResponse, ACMERequestHandler, ACMEResponse

# Naive regex just to make sure domains have basic validation
# before the server creates authorizations. We're not worried about
# clients submitting domains that don't resolve or submitting
# malformed domains, those will error in the challenge phase.
DOMAIN_REGEX = re.compile(r"^[^-][A-Za-z0-9-\.]+[^-]$")


class ACMEOrderResponse(ACMEResponse[OrderResource]):
    settings = inject.attr(ACMESettings)

    def __init__(self, *args, **kwargs):
        user_agent = kwargs.pop("user_agent", None)
        if user_agent is None:
            raise Exception("ACMEOrderResponse MUST include user_agent parameter")

        super().__init__(*args, **kwargs)

        if re.match(self.settings.mask_order_processing_status_ua_match, user_agent):
            self.msg._mask_processing_status = True

    def serialize(self) -> str:
        if self.msg:
            return self.msg.model_dump_json(indent=2)
        else:
            return ""


class NewOrderRequestHandler(ACMERequestHandler):
    """Handles new order requests and creates all supported authorizations and challenges."""

    message_type = NewOrderMessage
    settings = inject.attr(ACMESettings)
    challenge_service = inject.attr(IChallengeService)
    authorization_service = inject.attr(IAuthorizationService)
    order_service = inject.attr(IOrderService)
    directory_service = inject.attr(IDirectoryService)

    def process(self, msg: NewOrderMessage):
        self.identifier_validator(msg)
        # We ignore notBefore and notAfter. I know part of the spec is
        # that we MUST NOT issue a cert with contents other than those requested,
        # but this is an internal ACME server and we know better than the clients.
        # Also our CA overriddes several of our settings anyway when we forward the
        # finalized CSR, so good luck enforcing the RFC.
        resource_expiration = datetime.now() + self.settings.resource_expiration_delta

        new_order = OrderResource(
            account_id=msg.account.id,
            expires=resource_expiration,
            identifiers=msg.payload.identifiers,
            authorizations=[],
        )

        new_order = self.order_service.create(new_order)
        for identifier in new_order.identifiers:
            authorization = AuthorizationResource(
                order_id=new_order.id,
                expires=resource_expiration,
                identifier=identifier,
            )

            authz: AuthorizationResource = self.authorization_service.create(
                authorization
            )
            new_order.authorizations.append(authz)

            for chall_type in self.settings.challenges_available:
                # The spec doesn't say this has to be base64 encoded, but
                # certbot thinks it is?
                token = josepy.encode_b64jose(
                    "".join(
                        random.choices(string.ascii_letters + string.digits, k=50)
                    ).encode()
                )
                match chall_type:
                    case Challenges.http_01:
                        chall = HTTPChallengeResource(
                            authz_id=authz.id,
                            # The spec doesn't say this has to be base64 encoded, but
                            # certbot thinks it is?
                            token=token,
                        )
                        chall = self.challenge_service.create(chall)
                        authz.challenges.append(chall)
                    case Challenges.custom:
                        chall = CustomChallengeResource(authz_id=authz.id, token=token)
                        chall = self.challenge_service.create(chall)
                        authz.challenges.append(chall)
                    case _:
                        raise Exception(
                            f"No challenge mapped to challenge type {chall_type.value}"
                        )

        return ACMEOrderResponse(
            msg=new_order,
            code=201,
            location=self.directory_service.url_for(ACMEEndpoint.order, new_order.id),
            user_agent=self.headers.get("User-Agent", ""),
        )

    def identifier_validator(self, msg: NewOrderMessage):
        if len(msg.payload.identifiers) == 0:
            raise ACMEError(
                type="malformed", detail="No identifiers present in payload"
            )

        if len(msg.payload.identifiers) > self.settings.max_identifiers:
            raise ACMEError(
                type="malformed",
                detail=f"Too many identifiers. Max is {self.settings.max_identifiers}",
            )

        # All identifiers must be of type "dns" and a single well-formed domain.
        # wildcards MUST NOT be issued.
        for identifier in msg.payload.identifiers:
            if identifier.type != IdentifierType.dns:
                raise ACMEError(
                    type="rejectedIdentifier",
                    detail=f"{identifier.type} is not supported",
                )

            if "*" in identifier.value:
                raise ACMEError(
                    type="rejectedIdentifier",
                    detail="Wildcard identifiers are not supported",
                )

            if not DOMAIN_REGEX.match(identifier.value):
                raise ACMEError(
                    type="rejectedIdentifier",
                    detail=f"{identifier.value} is formatted incorrectly",
                )

            for domain_regex in self.settings.blacklisted_domains:
                if re.match(domain_regex, identifier.value):
                    raise ACMEError(
                        type="rejectedIdentifier",
                        detail=f"{identifier.value} is not allowed",
                    )


class OrderFinalizationRequestHandler(ACMERequestHandler):
    """Order finalization is performed when all authorizations are valid and the client
    wishes to submit a certificate signing request."""

    message_type = OrderFinalizationMessage
    order_service = inject.attr(IOrderService)
    directory_service = inject.attr(IDirectoryService)
    authorization_service = inject.attr(IAuthorizationService)

    def process(
        self,
        msg: OrderFinalizationMessage,
    ):
        order = self.order_service.resolve_state(msg.resource)

        if order.status != OrderStatus.ready:
            raise ACMEError(type="orderNotReady")

        # This probably never gets hit, the resolve_state method above should force orders
        # to an invalid state if they're expired.
        self.validate_expiration(order)
        self.validate_authorizations(order)

        csr = x509.load_der_x509_csr(josepy.decode_b64jose(msg.payload.csr))
        self.validate_identifiers(order, csr)

        order = self.order_service.process_finalization(order, csr)

        order_without_links = without(order, ["certificate", "finalize"])

        return ACMEOrderResponse(
            msg=order_without_links,
            code=200,
            user_agent=self.headers.get("User-Agent", ""),
        )

    def validate_expiration(self, order: OrderResource):
        if datetime.now() > order.expires:
            raise ACMEError(detail="Order has expired")

    def validate_authorizations(self, order: OrderResource):
        authz_status = [
            a.status == AuthorizationStatus.valid for a in order.authorizations
        ]

        if not all(authz_status):
            raise ACMEBadCSRError(ACMEBadCSRDetail.orderNotAuthorized)

    def validate_identifiers(
        self, order: OrderResource, csr: x509.CertificateSigningRequest
    ):
        cn: str = None
        try:
            cn: str = csr.subject.get_attributes_for_oid(x509.OID_COMMON_NAME)[0].value
        except IndexError:
            # Some clients don't pass a CN, we can use the first SAN in that case
            cn = None

        sans: list[str] = []

        try:
            sans_ext = csr.extensions.get_extension_for_oid(
                x509.OID_SUBJECT_ALTERNATIVE_NAME
            )
        except x509.ExtensionNotFound:
            raise ACMEBadCSRError(ACMEBadCSRDetail.sansRequired)

        for san in sans_ext.value:
            if not isinstance(san, x509.DNSName):
                raise ACMEBadCSRError(ACMEBadCSRDetail.invalidSan)
            sans.append(san.value)

        if cn is None:
            cn = sans[0]

        if cn not in sans:
            raise ACMEBadCSRError(ACMEBadCSRDetail.cnMissingFromSan)

        csr_domains = {cn} | set(sans)
        order_identifiers = set(i.value for i in order.identifiers)

        if csr_domains != order_identifiers:
            raise ACMEBadCSRError(ACMEBadCSRDetail.csrOrderMismatch)


class OrderRequestHandler(ACMERequestHandler):
    """Returns the order resource with links to all authorizations."""

    message_type = OrderMessage
    order_service = inject.attr(IOrderService)
    directory_service = inject.attr(IDirectoryService)
    authorization_service = inject.attr(IAuthorizationService)

    def process(
        self,
        msg: OrderMessage,
    ):
        order = self.order_service.resolve_state(msg.resource)

        return ACMEOrderResponse(
            msg=order,
            code=200,
            location=self.directory_service.url_for(ACMEEndpoint.order, order.id),
            user_agent=self.headers.get("User-Agent", ""),
        )



from acmev2.models import ChallengeResource
from acmev2.services import ACMEResourceType, ACMEEndpoint
from .base import SignedACMEMessage, AuthenticatedACMEMessage, WithResourceUrl


class ChallengeMessage(SignedACMEMessage, WithResourceUrl[ChallengeResource]):
    """https://datatracker.ietf.org/doc/html/rfc8555/#section-7.5.1"""

    protected_resource_type = ACMEResourceType.challenge
    protected_url_type = ACMEEndpoint.challenge



import abc
import random
import string
from enum import Enum
from typing import Any, Mapping

import inject
import josepy.jwk
import parse
from cryptography.x509 import CertificateSigningRequest

import acmev2.models as models
from acmev2.models import (
    CertResource,
    AuthorizationStatus,
    ChallengeStatus,
    OrderStatus,
)
from acmev2 import challenges
from datetime import datetime
import logging

logger = logging.getLogger(__name__)


class ACMEResourceType(str, Enum):
    order = "order"
    account = "acct"
    authz = "authz"
    challenge = "chall"
    cert = "cert"


class ACMEEndpoint(str, Enum):
    newNonce = "newNonce"
    newOrder = "newOrder"
    order = "order"
    newAccount = "newAcct"
    account = "acct"
    authz = "authz"
    challenge = "chall"
    finalize = "finalize"
    cert = "cert"


class INonceService(abc.ABC):

    @abc.abstractmethod
    def generate(self) -> str:
        """Generate a nonce for use with subsequent ACME requests.

        Returns:
            nonce (str)
        """

    @abc.abstractmethod
    def consume(self, nonce: str) -> bool:
        """Validate and consume the nonce passed in. The nonce must be made invalid for any
        future requests. If consume is called twice the second result must never be True.

        Args:
            nonce (str): The nonce.

        Returns:
            valid (bool): True if the nonce existed in storage and was able to be consumed.
        """


class IAccountService(abc.ABC):

    @abc.abstractmethod
    def create(
        self, resource: models.AccountResource, eab_kid: str = None
    ) -> models.AccountResource:
        """Persist an account resource and the associated jwk to the datastore. If eab_kid is passed in,
        also bind the external account."""

    @abc.abstractmethod
    def get(self, account_id: str) -> models.AccountResource | None:
        pass

    @abc.abstractmethod
    def get_by_jwk(self, jwk: josepy.jwk.JWK) -> models.AccountResource | None:
        pass

    @abc.abstractmethod
    def get_eab_hmac(self, account_id: str) -> str | None:
        pass

    @abc.abstractmethod
    def check_access(
        self, account_id: str, resource_id: str, resource_type: ACMEResourceType
    ) -> bool:
        """Returns True if the account can access the specified resource."""


class IAuthorizationService(abc.ABC):

    @abc.abstractmethod
    def create(
        self, resource: models.AuthorizationResource
    ) -> models.AuthorizationResource:
        """Persist an authorization resource to the datastore. The order passed in is the parent."""

    @abc.abstractmethod
    def update_status(
        self, authz: models.AuthorizationResource, new_state: AuthorizationStatus
    ) -> models.AuthorizationResource:
        pass

    @abc.abstractmethod
    def get(self, authz_id: str) -> models.AuthorizationResource:
        pass

    @abc.abstractmethod
    def get_by_order(self, order_id: str) -> list[models.AuthorizationResource]:
        pass

    @abc.abstractmethod
    def get_by_chall(self, chall_id: str) -> models.AuthorizationResource | None:
        pass


class IOrderService(abc.ABC):

    authorization_service = inject.attr(IAuthorizationService)

    @abc.abstractmethod
    def create(self, resource: models.OrderResource) -> models.OrderResource:
        """Persists an order resource to the datastore. The account passed in is the
        owner and initiator of the request."""

    @abc.abstractmethod
    def update_status(
        self, order: models.OrderResource, new_state: OrderStatus
    ) -> models.OrderResource:
        pass

    @abc.abstractmethod
    def get(self, order_id: str) -> models.OrderResource | None:
        pass

    @abc.abstractmethod
    def process_finalization(
        self, order: models.OrderResource, csr: CertificateSigningRequest
    ) -> models.OrderResource:
        """Processes the CSR asynchronously and updates the datastore order to reflect
        the current state of the order.
        """

    def resolve_state(self, order: models.OrderResource) -> models.OrderResource:
        """Checks and updates the state of an order and its children."""

        def try_transition_order_to(
            order: models.OrderResource, status: OrderStatus
        ) -> tuple[bool, models.OrderResource]:
            # If the order is in a terminal state we leave it alone
            if order.status in [OrderStatus.invalid, OrderStatus.valid]:
                return False, order

            if order.status != status:
                return True, self.update_status(order, status)

            return False, order

        def try_transition_auth_to(
            authz: models.AuthorizationResource, status: AuthorizationStatus
        ) -> tuple[bool, models.AuthorizationStatus]:
            # If the authorization is in a terminal state we leave it alone
            if authz.status in [
                AuthorizationStatus.invalid,
                AuthorizationStatus.revoked,
                AuthorizationStatus.deactivated,
                AuthorizationStatus.expired,
            ]:
                return False, authz

            if authz.status != status:
                return True, self.authorization_service.update_status(authz, status)

            return False, authz

        if order.expires < datetime.now():
            # If the order has expired and has not reached a terminal state, it becomes invalid
            _, order = try_transition_order_to(order, OrderStatus.invalid)

        for authz in order.authorizations:
            # If the authz has expired and has not reached a terminal state, it becomes expired
            if authz.expires < datetime.now():
                _, authz = try_transition_auth_to(authz, AuthorizationStatus.expired)
                _, order = try_transition_order_to(order, OrderStatus.invalid)

            if authz.status in [
                AuthorizationStatus.invalid,
                AuthorizationStatus.revoked,
                AuthorizationStatus.deactivated,
                AuthorizationStatus.expired,
            ]:
                _, order = try_transition_order_to(order, OrderStatus.invalid)

        if order.status == OrderStatus.pending:
            # Now check to see if all authorizations are valid
            if len(order.authorizations) > 0 and all(
                [a.status == AuthorizationStatus.valid for a in order.authorizations]
            ):

                _, order = try_transition_order_to(order, OrderStatus.ready)

        return order


class ICertService(abc.ABC):

    @abc.abstractmethod
    def get(self, ord_id: str) -> CertResource | None:
        """Get the certificate created during order finalization, or none if finalization isn't complete."""


class IChallengeService(abc.ABC):

    authz_service = inject.attr(IAuthorizationService)

    @abc.abstractmethod
    def create(
        self, authz_id: str, chall: models.ChallengeResource
    ) -> models.ChallengeResource:
        """Persist a challenge to the datastore. The authorization passed in is the parent."""

    @abc.abstractmethod
    def update_status(
        self, chall: models.ChallengeResource, new_state: ChallengeStatus
    ) -> models.ChallengeResource:
        pass

    @abc.abstractmethod
    def get(self, chall_id: str) -> models.ChallengeResource | None:
        pass

    @abc.abstractmethod
    def queue_validation(
        self, chall: models.ChallengeResource
    ) -> models.ChallengeResource:
        """Add the challenge to a background processing queue to perform verification at a later date."""

    def validate(
        self,
        acct_jwk: josepy.jwk.JWK,
        order: models.OrderResource,
        authz: models.AuthorizationResource,
        challenge: models.ChallengeResource,
    ) -> models.ChallengeResource:
        """Perform validation of the challenge by ensuring the client has ownership of the domain and persist
        changes to the authentication and challenge resources to the datastore."""

        # TODO: Add threadsafe and challenge status guards

        ChallengeValidator = challenges.validator_for(challenge.type)
        validator = ChallengeValidator(acct_jwk, authz, challenge)
        if validator.validate():
            order_service = inject.instance(IOrderService)
            # The authorization becomes valid
            self.authz_service.update_status(authz, AuthorizationStatus.valid)
            self.update_status(challenge, ChallengeStatus.valid)

            # Get the most up-to-date version of the order
            order = order_service.get(authz.order_id)
            order_service.resolve_state(order)

        return challenge


class IDirectoryService(abc.ABC):

    root_url = "https://change.me/acme"
    external_account_required = False

    url_templates = {
        ACMEEndpoint.newNonce: "newNonce",
        ACMEEndpoint.order: "order/{identifier}",
        ACMEEndpoint.newOrder: "newOrder",
        ACMEEndpoint.finalize: "order/{identifier}/finalize",
        ACMEEndpoint.newAccount: "newAcct",
        ACMEEndpoint.account: "acct/{identifier}",
        ACMEEndpoint.authz: "authz/{identifier}",
        ACMEEndpoint.cert: "cert/{identifier}",
        ACMEEndpoint.challenge: "chall/{identifier}",
    }

    def get_directory(self) -> Mapping[str, Any]:
        """Generates an ACME directory object with supported endpoints and metadata. The format of the
        directory is defined here: https://datatracker.ietf.org/doc/html/rfc8555/#section-7.1.1

        Returns:
            directory (Mapping[str, Any])
        """
        directory = {
            "newAccount": f"{self.root_url}/{self.url_templates[ACMEEndpoint.newAccount]}",
            "newNonce": f"{self.root_url}/{self.url_templates[ACMEEndpoint.newNonce]}",
            "newOrder": f"{self.root_url}/{self.url_templates[ACMEEndpoint.newOrder]}",
            "meta": {
                "externalAccountRequired": self.external_account_required,
                "termsOfService": self.tos_url() or f"{self.root_url}/terms-of-service",
                "website": self.website_url() or f"{self.root_url}",
            },
        }
        random_key = "".join(
            random.choices(string.ascii_letters + string.digits + "-_", k=10)
        )
        directory[random_key] = (
            "https://community.letsencrypt.org/t/adding-random-entries-to-the-directory/33417"
        )

        return directory

    def website_url(self) -> str:
        """Overridable website url returned in the directory meta."""
        pass

    def tos_url(self) -> str:
        """Overridable ToS url returned in the directory meta."""
        pass

    def identifier_from_url(self, endpoint: ACMEEndpoint, url: str) -> str | None:
        """Translates the url in the header of an ACME request to a resource id"""
        url_base = self.root_url
        if not url_base.endswith("/"):
            url_base += "/"

        if parse_result := parse.parse(url_base + self.url_templates[endpoint], url):
            identifier = parse_result.named.get("identifier")
            if not identifier:
                logger.warning(
                    "Unable to extract identifier from url: %s at endpoint %s",
                    url,
                    endpoint.value,
                )
            return parse_result.named.get("identifier")

    def url_for(self, endpoint: ACMEEndpoint, identifier: str = None) -> str:
        """Gets the url for an ACME resource"""
        url_base = self.root_url
        if not url_base.endswith("/"):
            url_base += "/"

        return url_base + self.url_templates[endpoint].format(identifier=identifier)



from acmev2.models import NewAccountRequestSchema
from .base import SignedACMEMessage


class NewAccountMessage(SignedACMEMessage[NewAccountRequestSchema]):
    """https://datatracker.ietf.org/doc/html/rfc8555/#section-7.3"""



from .services import (
    INonceService,
    IDirectoryService,
    IAccountService,
    IOrderService,
    IAuthorizationService,
    IChallengeService,
    ACMEEndpoint,
    ACMEResourceType,
    ICertService,
)



import inject
import josepy.jwk

from acmev2.models import AuthorizationResource, HTTPChallengeResource
from acmev2.settings import ACMESettings


class ValidationError(Exception):
    pass


class ChallengeValidator:

    settings = inject.attr(ACMESettings)

    def __init__(
        self,
        acct_jwk: josepy.jwk.JWK,
        authz: AuthorizationResource,
        chall: HTTPChallengeResource,
    ):
        self.chall = chall
        self.authz = authz
        self.acct_jwk = acct_jwk

    def validate(self) -> bool:
        return False



import inject
from acmev2.handlers import handle, ChallengeRequestHandler
from acmev2.challenges import validator_for
from acmev2.models import (
    OrderResource,
    AccountResource,
    ChallengeStatus,
    AuthorizationStatus,
)
from acmev2.tests.conftest import MessageBuilder
from acmev2.tests.helpers import authz_by_url
from acmev2.services import IChallengeService
from box import Box
import pytest
from xprocess import ProcessStarter
import tempfile
import os


class TestCertbot:

    @pytest.fixture
    def acme_server(self, xprocess):
        app_server = os.path.join(
            os.path.dirname(__file__), "../xprocess/acme_server.py"
        )

        class Starter(ProcessStarter):
            terminate_on_interrupt = True
            timeout = 5
            # startup pattern
            pattern = "Debugger PIN"

            # command to start process
            args = ["uv", "run", app_server]

        # ensure process is running
        xprocess.ensure("acme_server", Starter)

        yield

        # clean up whole process tree afterwards
        xprocess.getinfo("acme_server").terminate()

    @pytest.mark.slow
    def test_certbot(
        self,
        acme_server,
    ):
        import subprocess

        with tempfile.TemporaryDirectory() as cert_root:
            cmd = [
                "certbot",
                "--work-dir",
                cert_root,
                "--logs-dir",
                cert_root + "/logs",
                "--config-dir",
                cert_root + "/config",
                "certonly",
                "-d",
                "acme.localhost",
                "--server",
                "http://localhost:8080/directory",
                "--agree-tos",
                "-m",
                "usr@amce.localhost",
                "--standalone",
                "-n",
            ]

            res = subprocess.run(cmd, capture_output=True)
            assert res.returncode == 0
            assert os.path.exists(
                cert_root + "/config/live/acme.localhost/fullchain.pem"
            )



from .authorization import AuthorizationRequestHandler
from .base import ACMERequestHandler, ACMEModelResponse
from .challenge import ChallengeRequestHandler
from .new_account import NewAccountRequestHandler
from .new_nonce import NewNonceRequestHandler
from .order import (
    NewOrderRequestHandler,
    OrderFinalizationRequestHandler,
    OrderRequestHandler,
)
from .certificate import CertRequestHandler
from .handler import handle



from __future__ import annotations

import copy
import json
from typing import Any, Callable, Generator, Mapping, TypeVar

import josepy
from pydantic import (
    BaseModel,
    NaiveDatetime,
    PlainSerializer,
    computed_field,
    model_serializer,
)
from typing_extensions import Annotated


class ACMEResource(BaseModel):
    _exclude_fields: list[Callable[[], Generator[str, None, None]]] = []

    def __init_subclass__(cls, **kwargs):
        exclude_fields: list[Callable[[], Generator[str, None, None]]] = []
        for subclass in cls.mro():
            if "exclude_fields" in subclass.__dict__:
                exclude_fields.append(getattr(subclass, "exclude_fields"))

        def new_init(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._exclude_fields = exclude_fields

        cls.__init__ = new_init

        super().__init_subclass__(**kwargs)

    @model_serializer(mode="wrap")
    def serialize(self, handler):
        result = handler(self)
        for func in self._exclude_fields:
            for field in func(self):
                if field in result:
                    del result[field]

        return result


class JoseJsonSchema(ACMEResource):
    payload: str
    protected: str
    signature: str

    @computed_field
    @property
    def payload_decoded(self) -> Mapping[str, Any]:
        try:
            return json.loads(josepy.decode_b64jose(self.payload).decode())
        except json.JSONDecodeError:
            return {}

    @computed_field
    @property
    def protected_decoded(self) -> Mapping[str, Any]:
        try:
            return json.loads(josepy.decode_b64jose(self.protected).decode())
        except json.JSONDecodeError:
            return {}


class EmptyMessageSchema(ACMEResource):
    pass


RFC3339Date = Annotated[
    NaiveDatetime,
    PlainSerializer(
        lambda _datetime: _datetime.strftime("%Y-%m-%dT%H:%M:%SZ"), return_type=str
    ),
]

T = TypeVar("T", bound=ACMEResource)


def without(obj: T, attrs: list[str]) -> T:
    new_obj = copy.copy(obj)
    new_obj._exclude_fields = copy.copy(obj._exclude_fields)
    new_obj._exclude_fields.append(lambda *_: (a for a in attrs))

    return new_obj



from acmev2.models import NewOrderRequestSchema, OrderFinalizationRequestSchema
from acmev2.models import OrderResource
from acmev2.services import ACMEResourceType, ACMEEndpoint
from .base import SignedACMEMessage, AuthenticatedACMEMessage, WithResourceUrl


class NewOrderMessage(
    SignedACMEMessage[NewOrderRequestSchema], AuthenticatedACMEMessage
):
    """https://datatracker.ietf.org/doc/html/rfc8555/#section-7.4"""


class OrderFinalizationMessage(
    SignedACMEMessage[OrderFinalizationRequestSchema],
    WithResourceUrl[OrderResource],
):
    """https://datatracker.ietf.org/doc/html/rfc8555/#section-7.5"""

    protected_resource_type = ACMEResourceType.order
    protected_url_type = ACMEEndpoint.finalize


class OrderMessage(
    SignedACMEMessage[NewOrderRequestSchema],
    WithResourceUrl[OrderResource],
):
    """https://datatracker.ietf.org/doc/html/rfc8555/#section-7.4"""

    protected_resource_type = ACMEResourceType.order
    protected_url_type = ACMEEndpoint.order



from typing import Any, Callable, Mapping, TypedDict
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric import rsa, padding
import inject
import josepy
from cryptography import x509
from cryptography.x509.oid import NameOID
import pytest
from acmev2.handlers import handle, NewAccountRequestHandler

from acmev2.models import (
    OrderResource,
    OrderStatus,
    AuthorizationStatus,
    ChallengeStatus,
    NewAccountRequestSchema,
    AuthorizationResource,
)
from acmev2.handlers import ACMEModelResponse
from acmev2.services import (
    IDirectoryService,
    IChallengeService,
    ACMEEndpoint,
    IAuthorizationService,
)


def resp_has_error(resp: ACMEModelResponse, err: str, detail: str = None) -> bool:
    err_msg = resp.msg.model_dump()
    valid = err_msg.get("type") == err
    if detail:
        valid = valid and err_msg.get("detail") == detail

    return valid


def gen_private_key():
    return rsa.generate_private_key(public_exponent=65537, key_size=2048)


def gen_jwk_rsa(private_key: rsa.RSAPrivateKey):
    pem_private_key = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption(),
    )
    return josepy.JWKRSA.load(pem_private_key)


def gen_encoded_csr(*args, **kwargs) -> bytes:
    return josepy.encode_b64jose(
        gen_csr(*args, **kwargs).public_bytes(serialization.Encoding.DER)
    )


def gen_csr(
    key: rsa.RSAPrivateKey,
    cn: str = None,
    sans: list[str] = ["test.localhost", "test2.localhost"],
    order: OrderResource = None,
):
    if order:
        cn = order.identifiers[0].value
        sans = [order.identifiers[0].value] + [i.value for i in order.identifiers]

    builder = x509.CertificateSigningRequestBuilder().subject_name(
        x509.Name(
            [
                x509.NameAttribute(NameOID.COMMON_NAME, cn),
            ]
        )
    )

    if sans:
        builder = builder.add_extension(
            x509.SubjectAlternativeName(
                [
                    # Describe what sites we want this certificate for.
                    x509.DNSName(n)
                    for n in sans
                ]
            ),
            critical=False,
            # Sign the CSR with our private key.
        )
    csr = builder.sign(key, hashes.SHA256())

    return csr


def make_order_ready(order: OrderResource):
    order.status = OrderStatus.ready
    for authz in order.authorizations:
        authz.status = AuthorizationStatus.valid
        authz.challenges[0].status = ChallengeStatus.valid


def chall_by_url(url: str):
    directory_service = inject.instance(IDirectoryService)
    chall_service = inject.instance(IChallengeService)
    return chall_service.get(
        directory_service.identifier_from_url(ACMEEndpoint.challenge, url)
    )


def authz_by_url(url: str):
    directory_service = inject.instance(IDirectoryService)
    authz_service = inject.instance(IAuthorizationService)
    return authz_service.get(
        directory_service.identifier_from_url(ACMEEndpoint.authz, url)
    )


def validate_authz(authz: AuthorizationResource):
    authz.challenges[0].status = ChallengeStatus.valid
    authz.status = AuthorizationStatus.valid



import inject
from acmev2.handlers.base import ACMEModelResponse
from acmev2.models import AccountResource, OrderResource
from acmev2.handlers import (
    handle,
    ChallengeRequestHandler,
)
from acmev2.models.base import EmptyMessageSchema
from acmev2.services import IDirectoryService, ACMEEndpoint
from acmev2.tests.conftest import ChallengeRequest, MessageBuilder
from acmev2.tests.helpers import (
    authz_by_url,
    resp_has_error,
)
from box import Box
import josepy.jwk
from datetime import datetime, timedelta


class TestChallengeRequest:

    def test_begin_challenge_verification(
        self, default_order: OrderResource, chall_request: ChallengeRequest
    ):
        response: ACMEModelResponse = chall_request(EmptyMessageSchema())
        resource = Box(response.msg.model_dump())
        directory_service = inject.instance(IDirectoryService)
        assert (
            directory_service.url_for(
                ACMEEndpoint.authz, default_order.authorizations[0].id
            )
            in response.headers["Link"]
        )
        assert response.code == 200
        assert resource.status == "processing"
        assert resource.type == "http-01"
        assert len(resource.token) > 1
        assert resource.url is not None

    def test_expired_authz(
        self, default_order: OrderResource, chall_request: ChallengeRequest
    ):
        authz_resource = default_order.authorizations[0]
        authz_resource.expires = datetime.now() - timedelta(hours=1)
        response: ACMEModelResponse = chall_request(EmptyMessageSchema())
        assert resp_has_error(response, "urn:ietf:params:acme:error:malformed")

    def test_not_found(self, chall_request: ChallengeRequest):
        response: ACMEModelResponse = chall_request(EmptyMessageSchema(), url="invalid")
        assert response.code == 404

    def test_unauthorized(
        self,
        second_account: tuple[AccountResource, josepy.jwk.JWK],
        chall_request: ChallengeRequest,
    ):
        second_account, second_account_jwk = second_account
        response: ACMEModelResponse = chall_request(
            EmptyMessageSchema(), acct_id=second_account.id, jwk_rsa=second_account_jwk
        )
        assert response.code == 401



import logging

import inject
import pydantic

from acmev2.errors import (
    ACMEError,
    ACMEBadNonceError,
    ACMEInvalidContentTypeError,
    ACMEMethodNotAllowedError,
    ACMEUnauthorizedError,
)
from acmev2.handlers import ACMEModelResponse, ACMERequestHandler
from acmev2.logging import LazyLoggedResponse
from acmev2.handlers.base import HttpVerb
from acmev2.messages.base import (
    ACMEMessage,
    AuthenticatedACMEMessage,
    SignedACMEMessage,
)
from acmev2.models import AccountStatus, ProblemResource
from acmev2.services import INonceService

logger = logging.getLogger(__name__)


def format_problem(error: ACMEError):
    """Formats an ACMEError as a problem document response.

    Args:
        error (ACMEError): The error to transform.

    Returns:
        ACMEModelResponse: A model response that will serialize to a problem document.
    """
    resp = ACMEModelResponse(
        headers={"Content-Type": "application/problem+json"},
        msg=ProblemResource(type=error.type, detail=error.detail),
        code=error.code,
    )

    for link, rel in error.header_links:
        resp.add_link_header(link, rel)

    return resp


class LazyLoggedMessage:
    def __init__(self, message: ACMEMessage):
        self.message = message

    def __str__(self):
        import json

        if isinstance(self.message, SignedACMEMessage):
            return json.dumps(
                {
                    "payload": self.message.jose_message.payload_decoded,
                    "protected": self.message.jose_message.protected_decoded,
                },
                indent=2,
            )
        else:
            return json.dumps(self.message.body, indent=2)


def handle(req: ACMERequestHandler) -> ACMEModelResponse:
    """Handles the ACME request and any errors it raises, along with
    ensuring every request that needs a nonce has one returned.

    Args:
        req (ACMERequestHandler): The request handler to be executed.

    Returns:
        ACMEModelResponse: A serializable model that an ACME client can understand.
    """
    nonce_service = inject.instance(INonceService)
    resp: ACMEModelResponse = None
    acme_message: ACMEMessage | None = None
    try:
        if (
            req.msg is not None
            and req.verb == HttpVerb.POST
            and req.headers.get("Content-Type") != "application/jose+json"
        ):
            raise ACMEInvalidContentTypeError()

        if req.verb not in req.allowed_verbs:
            raise ACMEMethodNotAllowedError()

        acme_message = req.extract_message(req.msg)
        logger.debug("client message: %s", LazyLoggedMessage(acme_message))

        # Request url integrity: https://datatracker.ietf.org/doc/html/rfc8555/#section-6.4
        if isinstance(acme_message, SignedACMEMessage):
            if acme_message.url != req.request_url:
                raise ACMEUnauthorizedError()

        # Ensure account hasn't been revoked or deactivated
        if isinstance(acme_message, AuthenticatedACMEMessage):
            if acme_message.account.status != AccountStatus.valid:
                raise ACMEError(detail="Account not valid")

        # Consume the nonce right before processing the request
        if req.requires_nonce and not nonce_service.consume(acme_message.nonce):
            raise ACMEBadNonceError()

        resp = req.process(acme_message)

        if resp.content_type:
            resp.headers["Content-Type"] = resp.content_type
    except ACMEError as exc:
        # These are normal expected errors raised by handlers
        logging_extra_info = ""
        if acme_message and isinstance(acme_message, SignedACMEMessage):
            logging_extra_info += f" url<{acme_message.url}>"
        if acme_message and isinstance(acme_message, AuthenticatedACMEMessage):
            logging_extra_info += f" account<{acme_message.account_id}>"

        logging.info(
            f"ACMEError - type<{exc.type}> detail<{exc.detail}>{logging_extra_info}"
        )
        resp = format_problem(exc)
    except pydantic.ValidationError:
        # Pydantic errors usually happen when the request is malformed
        logger.exception("Error validating Pydantic model")
        resp = format_problem(ACMEError(type="malformed"))
    except Exception:
        # All other errors return a generic problem document
        logging.exception("Unhandled exception caught at error boundry")
        resp = format_problem(ACMEError(type="serverInternal", code=500))

    if req.verb == HttpVerb.POST and "Replay-Nonce" not in resp.headers:
        # Every POST needs a nonce returned. Head/Get requests should only
        # generate a nonce with the newNonce handler.
        resp.headers["Replay-Nonce"] = nonce_service.generate()

    logger.debug("response: %s", LazyLoggedResponse(resp))
    return resp



import io
import json
import logging.config
from typing import Any, Callable, Mapping, TypedDict

import acme.jws
import acme
import acme.client
import inject
import josepy
import pytest
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import rsa, padding

import requests.adapters
import urllib3

from acmev2.models import (
    AccountResource,
    OrderResource,
    NewAccountRequestSchema,
    NewOrderRequestSchema,
    EmptyMessageSchema,
    AuthorizationSchema,
    OrderFinalizationRequestSchema,
)
from acmev2.handlers import (
    NewNonceRequestHandler,
    NewAccountRequestHandler,
    NewOrderRequestHandler,
    AuthorizationRequestHandler,
    OrderFinalizationRequestHandler,
    ChallengeRequestHandler,
    OrderRequestHandler,
    CertRequestHandler,
)
from acmev2.handlers.base import ACMEModelResponse
from acmev2.handlers.handler import handle

from acmev2.services import (
    IDirectoryService,
    INonceService,
    IAccountService,
    IAuthorizationService,
    IOrderService,
    IChallengeService,
    ICertService,
)
from acmev2.services.services import ACMEEndpoint

from .test_services import (
    MemoryDirectoryService,
    MemoryNonceService,
    MemoryAccountService,
    MemoryAuthorizationService,
    MemoryOrderService,
    MemoryChallengeService,
    MemoryCertService,
)

from . import helpers
import logging

DEFAULT_RESOURCE_URL = "http://acme.localhost/acme/resource"

Message = TypedDict("Message", {"payload": str, "protected": str, "signature": str})
AcctId = str
OrdId = str
ResourceUrl = str
MessageBuilder = Callable[[Mapping[str, Any], josepy.JWKRSA, AcctId], Message]


def default_request_headers():
    return {"Content-Type": "application/jose+json"}


@pytest.fixture(autouse=True)
def setup():
    import dnsmock

    # I have no idea why these are necessary. If they don't exist then DNS resolution
    # will intermittently fail. Suspect it has something to do with Docker networking
    # and embedded DNS but it's not important enough to fix.
    dnsmock.bind_ip("acme.localhost", 443, "127.0.0.1")
    dnsmock.bind_ip("acme.localhost", 80, "127.0.0.1")
    from acmev2 import settings

    bindings = [
        (INonceService, MemoryNonceService()),
        (IDirectoryService, MemoryDirectoryService()),
        (IAccountService, MemoryAccountService()),
        (IOrderService, MemoryOrderService()),
        (IAuthorizationService, MemoryAuthorizationService()),
        (IChallengeService, MemoryChallengeService()),
        (ICertService, MemoryCertService()),
        (settings.ACMESettings, settings.ACMESettings()),
    ]
    inject.configure(
        lambda binder: [binder.bind(api, impl) for api, impl in bindings],
        bind_in_runtime=False,
        clear=True,
    )
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    handler = logging.StreamHandler()
    handler.setLevel(logging.INFO)
    # handler.setFormatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    logger.addHandler(handler)


@pytest.fixture()
def private_key() -> rsa.RSAPrivateKey:
    return helpers.gen_private_key()


@pytest.fixture()
def jwk_rsa(private_key: rsa.RSAPrivateKey) -> josepy.JWK:
    return helpers.gen_jwk_rsa(private_key)


@pytest.fixture()
def message_builder(jwk_rsa: josepy.JWKRSA) -> MessageBuilder:
    default_jwk_rsa = jwk_rsa

    def builder(
        payload: Mapping[str, Any] = {},
        jwk_rsa: josepy.JWKRSA = None,
        acct_id: AcctId = None,
        url: str = None,
        nonce: str = None,
    ) -> Message:
        if not jwk_rsa:
            jwk_rsa = default_jwk_rsa

        nonce_service = inject.instance(INonceService)
        directory_service = inject.instance(IDirectoryService)

        header = {
            "alg": josepy.RS256.name,
            "nonce": nonce if nonce else nonce_service.generate(),
            "url": url or DEFAULT_RESOURCE_URL,
        }

        if acct_id:
            if "/" not in acct_id:
                acct_id = directory_service.url_for(ACMEEndpoint.account, acct_id)
            header["kid"] = acct_id
        else:
            header["jwk"] = jwk_rsa.public_key().to_json()

        def e(msg: dict):
            return josepy.encode_b64jose(json.dumps(msg).encode())

        msg: Message = {"protected": e(header), "payload": e(payload)}
        msg["signature"] = josepy.encode_b64jose(
            jwk_rsa.key.sign(
                f"{msg['protected']}.{msg['payload']}".encode(),
                padding.PKCS1v15(),
                hashes.SHA256(),
            )
        )

        return msg

    return builder


NewAccountRequest = Callable[[NewAccountRequestSchema], ACMEModelResponse]


@pytest.fixture
def new_account_request(
    message_builder: MessageBuilder,
) -> NewAccountRequest:
    def wrapped(msg: NewAccountRequestSchema) -> ACMEModelResponse:
        return handle(
            NewAccountRequestHandler(
                DEFAULT_RESOURCE_URL,
                "POST",
                msg=message_builder(msg.model_dump()),
                headers=default_request_headers(),
            )
        )

    return wrapped


NewOrderRequest = Callable[
    [NewOrderRequestSchema, AcctId, ResourceUrl], ACMEModelResponse
]


@pytest.fixture
def new_order_request(
    message_builder: MessageBuilder,
    default_account: AccountResource,
    default_order: OrderResource,
) -> NewOrderRequest:
    def wrapped(
        msg: NewOrderRequestSchema,
        acct_id: AcctId | None = None,
        url: ResourceUrl | None = None,
        request_url: str | None = None,
        headers: Mapping[str, str] = None,
        message_builder_kwargs={},
    ) -> ACMEModelResponse:
        if not url:
            url = "http://acme.localhost/acme/newOrder"

        if not request_url:
            request_url = url

        return handle(
            NewOrderRequestHandler(
                request_url,
                "POST",
                msg=message_builder(
                    msg.model_dump(),
                    acct_id=acct_id if acct_id else default_account.id,
                    url=url,
                    **message_builder_kwargs,
                ),
                headers=headers or default_request_headers(),
            )
        )

    return wrapped


OrderRequest = Callable[[EmptyMessageSchema, AcctId, ResourceUrl], ACMEModelResponse]


@pytest.fixture
def order_request(
    message_builder: MessageBuilder,
    default_account: AccountResource,
    default_order: OrderResource,
) -> OrderRequest:
    def wrapped(
        msg: EmptyMessageSchema,
        acct_id: AcctId | None = None,
        url: ResourceUrl | None = None,
        message_builder_kwargs={},
    ) -> ACMEModelResponse:
        return handle(
            OrderRequestHandler(
                (
                    url
                    if url
                    else inject.instance(IDirectoryService).url_for(
                        ACMEEndpoint.order, default_order.id
                    )
                ),
                "POST",
                msg=message_builder(
                    msg.model_dump(),
                    acct_id=acct_id if acct_id else default_account.id,
                    url=(
                        url
                        if url
                        else inject.instance(IDirectoryService).url_for(
                            ACMEEndpoint.order, default_order.id
                        )
                    ),
                    **message_builder_kwargs,
                ),
                headers=default_request_headers(),
            )
        )

    return wrapped


AuthorizationRequest = Callable[
    [EmptyMessageSchema, AcctId, ResourceUrl], ACMEModelResponse
]


@pytest.fixture
def authorization_request(
    message_builder: MessageBuilder,
    default_account: AccountResource,
    default_order: OrderResource,
) -> AuthorizationRequest:
    def wrapped(msg: AuthorizationSchema, **kwargs) -> ACMEModelResponse:
        if "acct_id" not in kwargs:
            kwargs["acct_id"] = default_account.id
        if "url" not in kwargs:
            kwargs["url"] = inject.instance(IDirectoryService).url_for(
                ACMEEndpoint.authz, default_order.authorizations[0].id
            )
        request_url = kwargs.pop("request_url", kwargs["url"])

        return handle(
            AuthorizationRequestHandler(
                request_url,
                "POST",
                msg=message_builder(msg.model_dump(), **kwargs),
                headers=default_request_headers(),
            )
        )

    return wrapped


ChallengeRequest = Callable[
    [EmptyMessageSchema, AcctId, ResourceUrl], ACMEModelResponse
]


@pytest.fixture
def chall_request(
    message_builder: MessageBuilder,
    default_account: AccountResource,
    default_order: OrderResource,
) -> AuthorizationRequest:
    def wrapped(msg: EmptyMessageSchema, **kwargs) -> ACMEModelResponse:
        if "acct_id" not in kwargs:
            kwargs["acct_id"] = default_account.id
        chall = default_order.authorizations[0].challenges[0]
        if "url" not in kwargs:
            directory_service = inject.instance(IDirectoryService)
            kwargs["url"] = directory_service.url_for(ACMEEndpoint.challenge, chall.id)
        request_url = kwargs.pop("request_url", kwargs["url"])

        return handle(
            ChallengeRequestHandler(
                request_url,
                "POST",
                msg=message_builder(msg.model_dump(), **kwargs),
                headers=default_request_headers(),
            )
        )

    return wrapped


FinalizeRequest = Callable[
    [OrderFinalizationRequestSchema, AcctId, ResourceUrl], ACMEModelResponse
]


@pytest.fixture
def finalize_order_request(
    message_builder: MessageBuilder,
    default_account: AccountResource,
    default_order: OrderResource,
) -> AuthorizationRequest:
    def wrapped(msg: OrderFinalizationRequestSchema, **kwargs) -> ACMEModelResponse:
        if "acct_id" not in kwargs:
            kwargs["acct_id"] = default_account.id

        if "url" not in kwargs:
            directory_service = inject.instance(IDirectoryService)
            kwargs["url"] = directory_service.url_for(
                ACMEEndpoint.finalize, default_order.id
            )
        request_url = kwargs.pop("request_url", kwargs["url"])
        headers = kwargs.pop("headers", default_request_headers())

        return handle(
            OrderFinalizationRequestHandler(
                request_url,
                "POST",
                msg=message_builder(msg.model_dump(), **kwargs),
                headers=headers,
            )
        )

    return wrapped


EABMessageBuilder = Callable[[str, str, josepy.JWKRSA], Message]


@pytest.fixture()
def eab_message_builder() -> EABMessageBuilder:
    def builder(kid: str, hmac_key: str, outerjwk: josepy.JWKRSA) -> Message:
        outerjwk_json = json.dumps(outerjwk.public_key().to_partial_json()).encode()

        eab = acme.jws.JWS.sign(
            outerjwk_json,
            josepy.jwk.JWKOct(key=josepy.decode_b64jose(hmac_key)),
            josepy.jwa.HS256,
            None,
            DEFAULT_RESOURCE_URL,
            kid,
        )

        return eab.to_partial_json()

    return builder


@pytest.fixture()
def default_account(setup, message_builder: MessageBuilder) -> AccountResource:
    response = handle(
        NewAccountRequestHandler(
            DEFAULT_RESOURCE_URL,
            "POST",
            msg=message_builder({"termsOfServiceAgreed": True}),
            headers=default_request_headers(),
        )
    )

    return response.msg


GenAccount = Callable[[], tuple[AccountResource, josepy.JWK]]


@pytest.fixture()
def gen_account(message_builder: MessageBuilder) -> GenAccount:
    def wrapped() -> tuple[AccountResource, josepy.JWK]:
        jwk_rsa = helpers.gen_jwk_rsa(helpers.gen_private_key())

        msg = message_builder(
            {
                "termsOfServiceAgreed": True,
            },
            jwk_rsa,
        )
        response = handle(
            NewAccountRequestHandler(
                DEFAULT_RESOURCE_URL, "POST", msg=msg, headers=default_request_headers()
            )
        )

        return response.msg, jwk_rsa

    return wrapped


@pytest.fixture()
def second_account(setup, gen_account) -> tuple[AccountResource, josepy.JWK]:
    return gen_account()


@pytest.fixture()
def default_order(
    setup, default_account, message_builder: MessageBuilder
) -> OrderResource:
    directory_service = inject.instance(IDirectoryService)
    new_order = message_builder(
        {
            "identifiers": [
                {"type": "dns", "value": "acme.localhost"},
                {"type": "dns", "value": "test2.acme"},
            ]
        },
        acct_id=default_account.id,
    )
    response = handle(
        NewOrderRequestHandler(
            DEFAULT_RESOURCE_URL,
            "POST",
            msg=new_order,
            headers=default_request_headers(),
        )
    )

    return response.msg


class LocalACMEAdapter(requests.adapters.HTTPAdapter):

    def req_by_path(self, path):
        import re

        if path.startswith("/acme/newnonce/"):
            return NewNonceRequestHandler
        elif path.startswith("/acme/newaccount/"):
            return NewAccountRequestHandler
        elif path.startswith("/acme/neworders/"):
            return NewOrderRequestHandler
        elif path.startswith("/acme/authz/"):
            return AuthorizationRequestHandler
        elif re.match(r"/acme/order/[a-zA-Z0-9]+/finalize", path):
            return OrderFinalizationRequestHandler
        elif path.startswith("/acme/chall/"):
            return ChallengeRequestHandler
        elif path.startswith("/acme/order/"):
            return OrderRequestHandler
        elif path.startswith("/acme/cert/"):
            return CertRequestHandler

        raise Exception("Message not found for path " + path)

    def send(
        self, request, stream=False, timeout=None, verify=True, cert=None, proxies=None
    ):
        json_body = json.loads(request.body or "{}")
        req = self.req_by_path(request.path_url)(
            request.method, request.headers, json_body
        )
        resp: ACMEModelResponse = handle(req)

        if resp.msg:
            content = io.BytesIO(resp.serialize().encode())
        else:
            content = io.BytesIO()
        urllib_response = urllib3.HTTPResponse(
            content,
            resp.headers,
            resp.code,
            preload_content=False,
        )

        return self.build_response(
            request,
            urllib_response,
        )


@pytest.fixture
def acme_client(private_key: rsa.RSAPrivateKey):
    return gen_acme_client(private_key)


def gen_acme_client(rsa_key: rsa.RSAPrivateKey = None):
    directory = acme.client.messages.Directory.from_json(
        {
            "newAuthz": "http://acme.localhost/acme/new-authz/",
            "newNonce": "http://acme.localhost/acme/newnonce/",
            "newAccount": "http://acme.localhost/acme/newaccount/",
            "newOrder": "http://acme.localhost/acme/neworders/",
            "revokeCert": "http://acme.localhost/acme/revokecert/",
            "keyChange": "http://acme.localhost/acme/key-change/",
            "renewalInfo": "http://acme.localhost/acme/renewal-info/",
            "meta": {},
        }
    )
    if rsa_key is None:
        rsa_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)

    account_key = josepy.JWKRSA(key=rsa_key)

    net = acme.client.ClientNetwork(account_key, user_agent="simple_acme_dns/v2")
    net.session.mount("http://acme.localhost/acme/", LocalACMEAdapter())
    net.session.mount("http://localhost/acme/", LocalACMEAdapter())
    return acme.client.ClientV2(directory, net=net)



from enum import Enum

import inject
from pydantic import Field, computed_field, model_serializer

from .base import ACMEResource, RFC3339Date


class ChallengeStatus(str, Enum):
    pending = "pending"
    valid = "valid"
    invalid = "invalid"
    processing = "processing"


class ChallengeType(str, Enum):
    none = "none"
    custom = "custom"
    http_01 = "http-01"


class ChallengeResource(ACMEResource):
    id: str | None = Field(exclude=True, default=None)
    authz_id: str | None = Field(exclude=True, default=None)
    type: ChallengeType = ChallengeType.none
    token: str
    status: ChallengeStatus = ChallengeStatus.pending
    validated: RFC3339Date | None = None

    @computed_field
    @property
    def url(self) -> str:
        from acmev2.services import ACMEEndpoint, IDirectoryService

        return inject.instance(IDirectoryService).url_for(
            ACMEEndpoint.challenge, self.id
        )

    @model_serializer(mode="wrap")
    def serialize(self, handler):
        result = handler(self)
        if self.status != ChallengeStatus.valid:
            del result["validated"]

        return result


class CustomChallengeResource(ChallengeResource):
    type: ChallengeType = ChallengeType.custom


class HTTPChallengeResource(ChallengeResource):
    type: ChallengeType = ChallengeType.http_01


ChallengeResource = CustomChallengeResource | HTTPChallengeResource



import pytest
import inject

from acmev2.services import IDirectoryService
from acmev2.settings import ACMESettings


class TestServicesMixin:

    @pytest.fixture(autouse=True)
    def setup_services(self):
        self.directory_service = inject.instance(IDirectoryService)
        self.settings = inject.instance(ACMESettings)



import inject
from pydantic import ValidationError
import pytest
from acmev2.settings import ACMESettings
from acmev2.models import AccountResource, NewOrderRequestSchema, Identifier
from acmev2.handlers import NewOrderRequestHandler, handle
from acmev2.tests.conftest import MessageBuilder
from acmev2.tests.helpers import resp_has_error
from box import Box

from acmev2.tests.test_base import TestServicesMixin
from ..conftest import NewOrderRequest


class TestNewOrderRequest(TestServicesMixin):

    def test_invalid_wildcard(self, new_order_request: NewOrderRequest):
        response = new_order_request(
            NewOrderRequestSchema(
                identifiers=[Identifier(type="dns", value="*.acme.localhost")]
            )
        )

        assert resp_has_error(response, "urn:ietf:params:acme:error:rejectedIdentifier")

    def test_invalid_identifier_type(self, new_order_request: NewOrderRequest):
        with pytest.raises(ValidationError):
            new_order_request(
                NewOrderRequestSchema(
                    identifiers=[Identifier(type="invalid", value="acme.localhost")]
                )
            )

    def test_too_many_identifiers(self, new_order_request: NewOrderRequest):
        self.settings.max_identifiers = 5

        def build_message():
            return new_order_request(
                NewOrderRequestSchema(
                    identifiers=[
                        Identifier(type="dns", value=f"{idx}.acme.localhost")
                        for idx in range(20)
                    ]
                )
            )

        response = build_message()
        assert resp_has_error(response, "urn:ietf:params:acme:error:malformed")

        self.settings.max_identifiers = 100
        response = build_message()
        assert response.code == 201

    def test_no_identifiers(self, new_order_request: NewOrderRequest):
        response = new_order_request(NewOrderRequestSchema(identifiers=[]))
        assert resp_has_error(response, "urn:ietf:params:acme:error:malformed")

    def test_order_created(self, new_order_request: NewOrderRequest):
        response = new_order_request(
            NewOrderRequestSchema(
                identifiers=[Identifier(type="dns", value="acme.localhost")]
            )
        )

        assert "Location" in response.headers
        assert response.code == 201
        resource = Box(response.msg.model_dump())

        assert len(resource["identifiers"]) == 1
        assert resource.finalize is not None
        # some clients require this even though it's optional!
        assert "certificate" in resource
        assert resource.status == "pending"

    def test_blacklisted_domains(self, new_order_request: NewOrderRequest):
        settings = inject.instance(ACMESettings)

        def make_request(domains: list[str]):
            return new_order_request(
                NewOrderRequestSchema(
                    identifiers=[
                        Identifier(type="dns", value=domain) for domain in domains
                    ]
                )
            )

        domain_test_cases = [
            ([r".*?\.invalid\.acme"], ["domain.invalid.acme"]),
            ([r".*?\.invalid\.acme"], ["domain.valid.acme", "domain.invalid.acme"]),
            # Testing the nth rule is evaluated, not just the first
            (
                [r".*?\.unused\.acme", r".*?\.invalid\.acme"],
                ["domain.valid.acme", "d.invalid.acme"],
            ),
            # must end with .acme
            ([r".*(?<!(\.acme))$"], ["invalid.com"]),
            ([r".*(?<!(\.acme))$"], ["invalid.com"]),
            # must not be subdomain.protected.acme or protected.acme
            ([r"^(.*\.)?protected.acme"], ["error.protected.acme"]),
            ([r"^(.*\.)?protected.acme"], ["protected.acme"]),
        ]

        for blacklist, domains in domain_test_cases:
            # reset blacklist
            settings.blacklisted_domains = []
            # first request should succeed
            assert make_request(domains).code == 201
            # set blacklist
            settings.blacklisted_domains = blacklist
            # second request should fail
            assert resp_has_error(
                make_request(domains),
                "urn:ietf:params:acme:error:rejectedIdentifier",
            ), {"blacklist": blacklist, "domains": domains}

    def test_unauthorized(self, new_order_request: NewOrderRequest):
        response = new_order_request(
            NewOrderRequestSchema(
                identifiers=[Identifier(type="dns", value="acme.localhost")]
            ),
            "invalid-acct-id",
        )

        assert resp_has_error(
            response, "urn:ietf:params:acme:error:accountDoesNotExist"
        )



from typing import Any, Generic, Mapping, Self, TypeVar
import typing

import inject
import josepy
from acmev2.errors import (
    ACMEError,
    ACMEBadNonceError,
    ACMEMalformedError,
    ACMEResourceNotFoundError,
    ACMESignatureVerificationError,
    ACMEUnauthorizedError,
)
from acmev2.models import JoseJsonSchema, EmptyMessageSchema, AccountResource
from pydantic import ValidationError, BaseModel
from acmev2.services import (
    IAccountService,
    ACMEEndpoint,
    ACMEResourceType,
    IDirectoryService,
    IOrderService,
    IChallengeService,
    IAuthorizationService,
    ICertService,
)
import logging

logger = logging.getLogger(__name__)

R = TypeVar("R", bound=BaseModel)
T = TypeVar("T", bound=BaseModel)


class ACMEMessage:

    body: Mapping[str, Any] = None

    def __init__(self, body: Mapping[str, Any]):
        self.body = body

    @property
    def nonce(self) -> str | None:
        return None

    @classmethod
    def from_json(cls, body: Mapping[str, Any]) -> Self:
        return cls(body)


class SignedACMEMessage(ACMEMessage, Generic[T]):
    has_nonce: True
    schema: T = EmptyMessageSchema

    _payload: T = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Awkward way to set the schema with the generic type
        bases = typing.get_args(self.__orig_bases__[0])
        if len(bases) > 0:
            self.schema = bases[0]

    @classmethod
    def from_json(cls, body: Mapping[str, Any]) -> Self:
        msg = cls(body)
        # All signed messages must have a body with a header, payload, and signature
        try:
            JoseJsonSchema.model_validate(body)
        except ValidationError as exc:
            raise ACMEMalformedError(exc)

        if not msg.verify_signature():
            raise ACMESignatureVerificationError()

        # There must be a "nonce" entry in the protected header
        if "nonce" not in msg.jose_message.protected_decoded:
            raise ACMEBadNonceError()

        return msg

    @property
    def nonce(self) -> str | None:
        return self.jose_message.protected_decoded["nonce"]

    @property
    def url(self) -> str | None:
        return self.jose_message.protected_decoded["url"]

    @property
    def payload(self) -> T:
        if not self._payload:
            self._payload = self.schema(**self.jose_message.payload_decoded)

        return self._payload

    _jose_message: JoseJsonSchema = None

    @property
    def jose_message(self) -> JoseJsonSchema:
        if not self._jose_message:
            try:
                self._jose_message = JoseJsonSchema(**dict(self.body))
            except ValidationError as exc:
                # Error in the body format
                raise ACMEMalformedError(exc)

        return self._jose_message

    _jws: josepy.JWS = None

    @property
    def jws(self) -> josepy.JWS:
        if not self._jws:
            self._jws = josepy.JWS.from_json(self.body)

        return self._jws

    _jwk: josepy.JWK = None

    @property
    def has_embedded_jwk(self) -> bool:
        return "jwk" in self.jose_message.protected_decoded

    @property
    def has_kid(self) -> bool:
        return "kid" in self.jose_message.protected_decoded

    @property
    @inject.autoparams()
    def jwk(self, account_service: IAccountService) -> josepy.JWK:
        if not self._jwk:
            if self.has_embedded_jwk and self.has_kid:
                raise ACMEError(type="malformed")

            protected = self.jose_message.protected_decoded
            if self.has_kid:
                if isinstance(self, AuthenticatedACMEMessage):
                    jwk = self.account.jwk if self.account else None
                else:
                    jwk = account_service.get(protected["kid"]).jwk

                if not jwk:
                    raise ACMEError(type="accountDoesNotExist")

                self._jwk = jwk
            else:
                self._jwk = josepy.JWK.from_json(protected["jwk"])

        return self._jwk

    def verify_signature(self) -> bool:
        return self.jws.verify(self.jwk)


class AuthenticatedACMEMessage:
    @property
    def account_id(self: SignedACMEMessage | Self) -> str:
        return self.jose_message.protected_decoded["kid"]

    _account: AccountResource | None = None

    @property
    def account(self: SignedACMEMessage | Self) -> AccountResource | None:
        if not self._account:
            auth_service = inject.instance(IAccountService)
            directory_service = inject.instance(IDirectoryService)
            account_identifier = directory_service.identifier_from_url(
                ACMEEndpoint.account, self.account_id
            )
            if not account_identifier:
                logger.error(
                    "Failed to extract account identifier from id: %s", self.account_id
                )

            self._account = auth_service.get(account_identifier)

        return self._account


class WithResourceUrl(Generic[R], AuthenticatedACMEMessage):

    protected_resource_type: ACMEResourceType
    protected_url_type: ACMEEndpoint

    @property
    def resource(self: SignedACMEMessage | Self) -> R:
        directory_service = inject.instance(IDirectoryService)
        account_service = inject.instance(IAccountService)

        resource_id = directory_service.identifier_from_url(
            self.protected_url_type, self.url
        )

        if not resource_id:
            raise ACMEResourceNotFoundError()

        concrete_resource: R = None
        match self.protected_resource_type:
            case ACMEResourceType.order:
                concrete_resource = inject.instance(IOrderService).get(resource_id)
            case ACMEResourceType.authz:
                concrete_resource = inject.instance(IAuthorizationService).get(
                    resource_id
                )
            case ACMEResourceType.challenge:
                concrete_resource = inject.instance(IChallengeService).get(resource_id)
            case ACMEResourceType.cert:
                concrete_resource = inject.instance(ICertService).get(resource_id)
            case _:
                raise Exception(
                    f"Could not get resource from type {self.protected_resource_type}"
                )

        if not concrete_resource:
            raise ACMEResourceNotFoundError()

        if not account_service.check_access(
            self.account.id, resource_id, self.protected_resource_type
        ):
            raise ACMEUnauthorizedError()

        return concrete_resource



from typing import Any, Generic, Mapping, Optional, Type, TypeVar
from urllib.parse import quote

import inject

from acmev2.messages.base import ACMEMessage
from acmev2.models import ACMEResource
from acmev2.services import IDirectoryService
import logging

logger = logging.getLogger(__name__)


class HttpVerb:
    HEAD = "HEAD"
    GET = "GET"
    POST = "POST"


class ACMERequestHandler:
    allowed_verbs = [HttpVerb.POST]
    message_type: Type[ACMEMessage] = ACMEMessage
    requires_nonce = True

    def __init__(
        self,
        request_url: str,
        verb: HttpVerb,
        headers: Mapping[str, Any] = None,
        msg: Optional[Mapping[str, Any]] = None,
    ):
        self.request_url = request_url
        self.verb = verb
        self.headers = headers or {}
        self.msg = msg

    def extract_message(self, msg: Optional[Mapping[str, Any]]):
        return self.message_type.from_json(msg)

    def process(msg: ACMEMessage) -> "ACMEModelResponse":
        pass


T = TypeVar("T")


class ACMEResponse(Generic[T]):
    content_type: str = "application/json"
    directory_service = inject.attr(IDirectoryService)
    msg: T | None

    def __init__(
        self,
        headers: Mapping[str, Any] = None,
        msg: Optional[T] = None,
        code: int = 200,
        location: str = None,
    ):
        self.headers = headers or {}
        self.msg = msg or None
        self.code = code

        if location:
            self.headers["Location"] = location

        directory_root = self.directory_service.root_url
        self.add_link_header(directory_root, "index")

    def add_link_header(self, link: str, rel: str):
        link_header = self.headers.get("Link", "")

        header_value = f'<{quote(link, safe="/:")}>;rel="{rel}"'
        if link_header:
            link_header += ","

        self.headers["Link"] = link_header + header_value


class ACMEModelResponse(ACMEResponse[ACMEResource]):
    def serialize(self) -> str:
        if self.msg:
            return self.msg.model_dump_json(indent=2)
        else:
            return ""



from time import sleep
from typing import Mapping
from cryptography.x509 import CertificateSigningRequest
from flask import Flask, Blueprint, jsonify, request, Response
import os
import inject
import parse
import datetime
import subprocess
import threading
from acmev2.models import ChallengeStatus, AuthorizationStatus, OrderStatus
from acmev2.models.challenge import ChallengeResource
from acmev2.models.order import OrderResource
from acmev2.handlers import handle as process_acme_request
from acmev2 import settings
from acmev2.handlers import ACMERequestHandler
from acmev2.handlers import (
    NewNonceRequestHandler,
    CertRequestHandler,
    OrderRequestHandler,
    NewOrderRequestHandler,
    ChallengeRequestHandler,
    NewAccountRequestHandler,
    AuthorizationRequestHandler,
    OrderFinalizationRequestHandler,
)
from acmev2.services.services import (
    ACMEEndpoint,
    INonceService,
    IDirectoryService,
    IAccountService,
    IOrderService,
    IAuthorizationService,
    IChallengeService,
    ICertService,
)
from acmev2.tests.test_services import (
    MemoryNonceService,
    MemoryAccountService,
    MemoryAuthorizationService,
    MemoryCertService,
    MemoryChallengeService,
    MemoryDirectoryService,
    MemoryOrderService,
)

DIR = os.path.dirname(__file__)

app = Flask(__name__)


class DirectoryService(IDirectoryService):
    root_url = "http://localhost:8080/acme"
    external_account_required = False


class ChallengeService(MemoryChallengeService):
    auth_service = inject.attr(IAuthorizationService)
    order_service = inject.attr(IOrderService)
    account_service = inject.attr(IAccountService)
    chall_service = inject.attr(IChallengeService)

    def queue_validation(self, chall: ChallengeResource) -> ChallengeResource:
        if chall.status == ChallengeStatus.processing:
            return chall

        super().queue_validation(chall)

        def _validate():
            sleep(1)
            authz = self.auth_service.get(chall.authz_id)
            order = self.order_service.get(authz.order_id)
            acct = self.account_service.get(order.account_id)
            # chall.status = ChallengeStatus.valid
            # authz.status = AuthorizationStatus.valid
            self.validate(acct.jwk, order, authz, chall)

            # self.chall_service.update_status(chall, ChallengeStatus.valid)
            # self.auth_service.update_status(authz, AuthorizationStatus.valid)

        threading.Thread(target=_validate).start()
        return chall


class OrderService(MemoryOrderService):
    def process_finalization(
        self, order: OrderResource, csr: CertificateSigningRequest
    ) -> OrderResource:
        order.status = OrderStatus.valid
        self.order_csrs[order.id] = csr
        return order


bindings = [
    (INonceService, MemoryNonceService()),
    (IDirectoryService, DirectoryService()),
    (IAccountService, MemoryAccountService()),
    (IOrderService, OrderService()),
    (IAuthorizationService, MemoryAuthorizationService()),
    (IChallengeService, ChallengeService()),
    (ICertService, MemoryCertService()),
    (settings.ACMESettings, settings.ACMESettings()),
]


inject.configure(
    lambda binder: [binder.bind(api, impl) for api, impl in bindings],
    bind_in_runtime=False,
    clear=True,
)


@app.route("/directory")
def directory():
    directory_service = inject.instance(IDirectoryService)
    return jsonify(directory_service.get_directory())


def translateACMERequest(ACMERequestType: ACMERequestHandler) -> Response:
    req = ACMERequestType(
        request.url,
        request.method,
        request.headers,
        request.json if request.is_json else None,
    )

    acme_resp = process_acme_request(req)
    resp_body = acme_resp.serialize() if acme_resp.msg else None
    resp = Response(
        resp_body,
        status=acme_resp.code,
        headers=acme_resp.headers,
    )

    return resp


@app.route("/acme/newNonce", methods=["HEAD", "GET"])
def newNonce():
    return translateACMERequest(NewNonceRequestHandler)


@app.route(
    "/acme/newAcct",
    methods=["POST"],
)
def newAccount():
    return translateACMERequest(NewAccountRequestHandler)


@app.route(
    "/acme/newOrder",
    methods=["POST"],
)
def newOrder():
    return translateACMERequest(NewOrderRequestHandler)


@app.route(
    "/acme/authz/<identifier>",
    methods=["POST"],
)
def authz(identifier):
    return translateACMERequest(AuthorizationRequestHandler)


@app.route(
    "/acme/chall/<identifier>",
    methods=["POST"],
)
def chall(identifier):
    return translateACMERequest(ChallengeRequestHandler)


@app.route(
    "/acme/order/<identifier>/finalize",
    methods=["POST"],
)
def finalize(identifier):
    return translateACMERequest(OrderFinalizationRequestHandler)


@app.route(
    "/acme/order/<identifier>",
    methods=["POST"],
)
def order(identifier):
    return translateACMERequest(OrderRequestHandler)


@app.route(
    "/acme/cert/<identifier>",
    methods=["POST"],
)
def cert(identifier):
    body = translateACMERequest(CertRequestHandler)
    return body


if __name__ == "__main__":
    inject.instance(IAccountService).hmacs["kid"] = "hmac"
    cert_path = "/tmp/acmev2.flask.crt"
    key_path = "/tmp/acmev2.flask.key"
    if not os.path.exists(key_path):
        subprocess.call(
            [
                "openssl",
                "req",
                "-new",
                "-newkey",
                "rsa:2048",
                "-days",
                "3650",
                "-nodes",
                "-x509",
                "-keyout",
                key_path,
                "-out",
                cert_path,
                "-subj",
                "/CN=acmeca.localhost",
                "-addext",
                "subjectAltName = DNS:acmeca.localhost",
            ]
        )

    app.run(
        debug=True, host="0.0.0.0", port=8080
    )  # , ssl_context=(cert_path, key_path))



from .base import JoseJsonSchema, EmptyMessageSchema, ACMEResource, without
from .identifiers import Identifier, IdentifierType
from .nonce import NewNonceRequestSchema
from .account import NewAccountRequestSchema, AccountResource, AccountStatus
from .problem import ProblemResource
from .authorization import (
    AuthorizationResource,
    AuthorizationStatus,
    AuthorizationSchema,
)
from .order import (
    NewOrderRequestSchema,
    OrderResource,
    OrderFinalizationRequestSchema,
    OrderStatus,
)

from .challenge import (
    HTTPChallengeResource,
    CustomChallengeResource,
    ChallengeResource,
    ChallengeStatus,
    ChallengeType,
)
from .certificate import CertResource



import inject
from acmev2.handlers import handle, ChallengeRequestHandler
from acmev2.challenges import validator_for
from acmev2.models import (
    OrderResource,
    AccountResource,
    ChallengeStatus,
    AuthorizationStatus,
)
from acmev2.tests.conftest import MessageBuilder
from acmev2.tests.helpers import authz_by_url
from acmev2.services import IChallengeService
from box import Box
import pytest
from xprocess import ProcessStarter
import tempfile
import os


class TestGetSSL:

    GETSSL_CONFIG = """
# vim: filetype=sh
#
# This file is read first and is common to all domains
#
# Uncomment and modify any variables you need
# see https://github.com/srvrco/getssl/wiki/Config-variables for details
#
# The staging server is best for testing (hence set as default)
SKIP_HTTP_TOKEN_CHECK=true
CA="http://acme.localhost:8080"
CHECK_REMOTE="false"
# This server issues full certificates, however has rate limits
#CA="https://acme-v02.api.letsencrypt.org"
SANS="acme.localhost"
# The agreement that must be signed with the CA, if not defined the default agreement will be used
#AGREEMENT=""

# Set an email address associated with your account - generally set at account level rather than domain.
#ACCOUNT_EMAIL="usr@acme.localhost"
ACCOUNT_KEY_LENGTH=4096
ACCOUNT_KEY="/home/vscode/.getssl/account.key"

# Account key and private key types - can be rsa, prime256v1, secp384r1 or secp521r1
#ACCOUNT_KEY_TYPE="rsa"
PRIVATE_KEY_ALG="rsa"
#REUSE_PRIVATE_KEY="true"

# Preferred Chain - use an different certificate root from the default
# This uses wildcard matching so requesting "X1" returns the correct certificate - may need to escape characters
# Staging options are: "(STAGING) Doctored Durian Root CA X3" and "(STAGING) Pretend Pear X1"
# Production options are: "ISRG Root X1" and "ISRG Root X2"
#PREFERRED_CHAIN="\(STAGING\) Pretend Pear X1"

# Uncomment this if you need the full chain file to include the root certificate (Java keystores, Nutanix Prism)
#FULL_CHAIN_INCLUDE_ROOT="true"

# The command needed to reload apache / nginx or whatever you use.
# Several (ssh) commands may be given using a bash array:
# RELOAD_CMD=('ssh:sshuserid@server5:systemctl reload httpd' 'logger getssl for server5 efficient.')
#RELOAD_CMD=""

ACL=('/tmp/acme/.well-known/acme-challenge')
USE_SINGLE_ACL="true"

# The time period within which you want to allow renewal of a certificate
#  this prevents hitting some of the rate limits.
# Creating a file called FORCE_RENEWAL in the domain directory allows one-off overrides
# of this setting
RENEW_ALLOW="30"

# Define the server type. This can be https, ftp, ftpi, imap, imaps, pop3, pop3s, smtp,
# smtps_deprecated, smtps, smtp_submission, xmpp, xmpps, ldaps or a port number which
# will be checked for certificate expiry and also will be checked after
# an update to confirm correct certificate is running (if CHECK_REMOTE) is set to true
SERVER_TYPE="https"
"""

    @pytest.fixture
    def acme_server(self, xprocess):
        app_server = os.path.join(
            os.path.dirname(__file__), "../xprocess/acme_server.py"
        )

        class Starter(ProcessStarter):
            terminate_on_interrupt = True
            timeout = 5
            # startup pattern
            pattern = "Debugger PIN"

            # command to start process
            args = ["uv", "run", app_server]

        # ensure process is running
        xprocess.ensure("acme_server", Starter)

        yield

        # clean up whole process tree afterwards
        xprocess.getinfo("acme_server").terminate()

    @pytest.fixture
    def chall_server_root(self, xprocess):
        class Starter(ProcessStarter):
            terminate_on_interrupt = True
            timeout = 5
            # startup pattern
            pattern = "Serving HTTP"

            # command to start process
            args = ["python3", "-m", "http.server", "-d", "/tmp/acme", "80"]

        # ensure process is running and return its logfile
        xprocess.ensure("chall_server", Starter)

        yield

        # clean up whole process tree afterwards
        xprocess.getinfo("chall_server").terminate()

    @pytest.mark.slow
    def _test_getssl(self, acme_server, chall_server_root):
        """This test won't work until getssl releases a fix that sees 'processing' and 'pending' as valid challenge responses"""
        import subprocess

        download_getssl = [
            "curl",
            "--silent",
            "--user-agent",
            "getssl/manual",
            "https://raw.githubusercontent.com/srvrco/getssl/latest/getssl",
            "--output",
            "getssl",
        ]

        getssl_path = "/tmp/getssl"
        if not os.path.exists(getssl_path):
            os.mkdir("/home/vscode/.getssl")
            assert subprocess.run(download_getssl, cwd="/tmp").returncode == 0
            subprocess.run(["chmod", "+x", "/tmp/getssl"])
            subprocess.run(["/tmp/getssl", "-c", "acme.localhost"])
            with open("/home/vscode/.getssl/getssl.cfg", "w") as cfg:
                cfg.write(self.GETSSL_CONFIG)

            os.mkdir("/home/vscode/.getssl/acme.localhost")
            with open("/home/vscode/.getssl/acme.localhost/getssl.cfg") as cfg:
                cfg.write("")

        cmd = ["/tmp/getssl", "acme.localhost", "-U", "-f"]

        res = subprocess.run(cmd, capture_output=True)

        assert res.returncode == 0
        assert os.path.exists("/home/vscode/.getssl/acme.localhost/acme.localhost.crt")



import inject

from acmev2.messages import (
    CertMessage,
)
from acmev2.services import (
    IOrderService,
    IAuthorizationService,
    IDirectoryService,
)
from acmev2.models import CertResource
from .base import ACMERequestHandler, ACMEResponse


class ACMECertResponse(ACMEResponse[CertResource]):
    content_type = "application/pem-certificate-chain"

    def serialize(self) -> str:
        # TODO: return different certs depending on the
        # client Accept header.
        #  application/pkix-cert - DER
        #  application/pkcs7-mime - DER
        return self.msg.pem


class CertRequestHandler(ACMERequestHandler):
    """Returns a certificate chain for the requested identifier. Only supports
    PEM-encoded certificates."""

    message_type = CertMessage
    order_service = inject.attr(IOrderService)
    directory_service = inject.attr(IDirectoryService)
    authorization_service = inject.attr(IAuthorizationService)

    def process(
        self,
        msg: CertMessage,
    ):
        # todo: this is actually multiple certs, change the response
        cert = msg.resource

        return ACMECertResponse(msg=cert, code=200)



from enum import Enum

from pydantic import Field

from .base import ACMEResource


class IdentifierType(str, Enum):
    dns = "dns"


class Identifier(ACMEResource):
    id: int | None = Field(exclude=True, default=None)
    type: IdentifierType
    value: str



from acmev2.models import AccountResource, AccountStatus
from acmev2.handlers import ACMEModelResponse
from acmev2.models import Identifier, NewOrderRequestSchema

from acmev2.tests.conftest import NewOrderRequest
from acmev2.tests.helpers import resp_has_error


class TestInvalidAccount:
    def test_deactivated_account(
        self, new_order_request: NewOrderRequest, default_account: AccountResource
    ):
        def make_request() -> ACMEModelResponse:
            return new_order_request(
                NewOrderRequestSchema(
                    identifiers=[Identifier(type="dns", value="acme.localhost")]
                )
            )

        resp = make_request()
        assert resp.code == 201

        default_account.status = AccountStatus.deactivated
        resp = make_request()
        assert resp_has_error(resp, "urn:ietf:params:acme:error:malformed")

    def test_revoked_account(
        self,
        default_account: AccountResource,
        new_order_request: NewOrderRequest,
    ):
        def make_request() -> ACMEModelResponse:
            return new_order_request(
                NewOrderRequestSchema(
                    identifiers=[Identifier(type="dns", value="acme.localhost")]
                )
            )

        resp = make_request()
        assert resp.code == 201

        default_account.status = AccountStatus.revoked
        resp = make_request()
        assert resp_has_error(resp, "urn:ietf:params:acme:error:malformed")



from enum import Enum


ERROR_PREFIX = "urn:ietf:params:acme:error:"

ERROR_CODES = {
    "accountDoesNotExist": "The request specified an account that does not exist",
    "alreadyRevoked": "The request specified a certificate to be revoked that has"
    " already been revoked",
    "badCSR": "The CSR is unacceptable (e.g., due to a short key)",
    "badNonce": "The client sent an unacceptable anti-replay nonce",
    "badPublicKey": "The JWS was signed by a public key the server does not support",
    "badRevocationReason": "The revocation reason provided is not allowed by the server",
    "badSignatureAlgorithm": "The JWS was signed with an algorithm the server does not support",
    "caa": "Certification Authority Authorization (CAA) records forbid the CA from issuing"
    " a certificate",
    "compound": 'Specific error conditions are indicated in the "subproblems" array',
    "connection": ("The server could not connect to the client to verify the domain"),
    "dns": "There was a problem with a DNS query during identifier validation",
    "dnssec": "The server could not validate a DNSSEC signed domain",
    "incorrectResponse": "Response received didn't match the challenge's requirements",
    "invalidContact": "The provided contact URI was invalid",
    "malformed": "The request message was malformed",
    "rejectedIdentifier": "The server will not issue certificates for the identifier",
    "orderNotReady": "The request attempted to finalize an order that is not ready to be finalized",
    "rateLimited": "There were too many requests of a given type",
    "serverInternal": "The server experienced an internal error",
    "tls": "The server experienced a TLS error during domain verification",
    "unauthorized": "The client lacks sufficient authorization",
    "unsupportedContact": "A contact URL for an account used an unsupported protocol scheme",
    "unknownHost": "The server could not resolve a domain name",
    "unsupportedIdentifier": "An identifier is of an unsupported type",
    "externalAccountRequired": "The server requires external account binding",
    "userActionRequired": "User action required",
}


class ACMEError(Exception):
    type: str = "malformed"
    detail: str = None
    code: int = 403
    header_links: list[tuple[str, str]] = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args)

        if "type" in kwargs:
            self.type = kwargs["type"]

        if "detail" in kwargs:
            self.detail = kwargs["detail"]

        if "code" in kwargs:
            self.code = kwargs["code"]

        if "header_links" in kwargs:
            self.header_links = kwargs["header_links"]

        if self.type in ERROR_CODES:
            self.type = f"{ERROR_PREFIX}{self.type}"


class ACMEBadNonceError(ACMEError):
    type = "badNonce"
    code = 400


class ACMEUnauthorizedError(ACMEError):
    type = "unauthorized"
    code = 401


class ACMEResourceNotFoundError(ACMEError):
    type = "resourceNotFound"
    code = 404


class ACMEMalformedError(ACMEError):
    type = "malformed"
    detail = "Malformed request"
    code = 400


class ACMESignatureVerificationError(ACMEError):
    type = "malformed"
    detail = "Malformed signature"
    code = 400


class ACMEToSAgreementError(ACMEError):
    type = "userActionRequired"
    code = 403
    detail = "Please read and agree to the terms of service"


class ACMEUserActionRequiredError(ACMEError):
    type = "userActionRequired"
    code = 403

    def __init__(self, detail: str, *args, **kwargs):
        self.detail = detail
        super().__init__(*args, **kwargs)


class ACMEMethodNotAllowedError(ACMEError):
    type = "malformed"
    code = 403


class ACMEBadCSRDetail(str, Enum):
    sansRequired = "SANS are required"
    invalidSan = "Invalid subjectAlternateName extension value"
    cnMissingFromSan = (
        "The common name must also appear in the subjectAlternateName extension"
    )
    csrOrderMismatch = (
        "The CSR must indiciate the exact same set of identifiers as the order"
    )
    orderNotAuthorized = "Order has not completed all authorizations"


class ACMEBadCSRError(ACMEError):
    type = "badCSR"

    def __init__(self, detail: ACMEBadCSRDetail):
        self.detail = str(detail)
        super().__init__()


class ACMEInvalidContentTypeError(ACMEError):
    code = 415


class JoseJsonFormatError(ACMEError):
    pass



from enum import Enum
from typing import Optional

from pydantic import BaseModel


class NewAccountRequestSchema(BaseModel):
    termsOfServiceAgreed: Optional[bool] = False
    contact: Optional[list[str]] = None
    onlyReturnExisting: Optional[bool] = False


class AccountStatus(str, Enum):
    valid = "valid"
    deactivated = "deactivated"
    revoked = "revoked"


class ProblemResource(BaseModel):
    type: str
    detail: Optional[str] = None



import inject
from acmev2.handlers import handle, ChallengeRequestHandler
from acmev2.challenges import validator_for
from acmev2.models import (
    OrderResource,
    AccountResource,
    ChallengeStatus,
    AuthorizationStatus,
)
from acmev2.tests.conftest import MessageBuilder, default_request_headers
from acmev2.tests.helpers import authz_by_url
from acmev2.services import IChallengeService
from box import Box
import pytest
from xprocess import ProcessStarter
import tempfile
import os
import josepy


class TestChallengeRequest:

    @pytest.fixture
    def chall_server_root(self, xprocess):
        with tempfile.TemporaryDirectory() as srv_root:

            class Starter(ProcessStarter):
                timeout = 5
                terminate_on_interrupt = True
                # startup pattern
                pattern = "Serving HTTP"

                # command to start process
                args = ["python3", "-m", "http.server", "-d", srv_root, "80"]

            # ensure process is running and return its logfile
            xprocess.ensure("chall_server", Starter)

            yield srv_root

            # clean up whole process tree afterwards
            xprocess.getinfo("chall_server").terminate(timeout=2)

    @pytest.fixture
    def https_chall_server_root(self, xprocess):
        https_server = os.path.join(
            os.path.dirname(__file__), "../xprocess/https_server.py"
        )
        with tempfile.TemporaryDirectory() as srv_root:

            class Starter(ProcessStarter):
                terminate_on_interrupt = True
                timeout = 5
                # startup pattern
                pattern = "Serving HTTPS"

                # command to start process
                args = ["uv", "run", https_server, "-d", srv_root]

            # ensure process is running and return its logfile
            xprocess.ensure("https_chall_server", Starter)

            yield srv_root

            # clean up whole process tree afterwards
            xprocess.getinfo("https_chall_server").terminate(timeout=2)

    @pytest.fixture
    def http_to_https_redirector(self, xprocess):
        http_server = os.path.join(
            os.path.dirname(__file__), "../xprocess/http_redirector.py"
        )

        class Starter(ProcessStarter):
            terminate_on_interrupt = True
            timeout = 5
            # startup pattern
            pattern = "Serving HTTP->HTTPS redirector"

            # command to start process
            args = ["uv", "run", http_server]

        # ensure process is running and return its logfile
        xprocess.ensure("redirector", Starter)

        yield

        # clean up whole process tree afterwards
        xprocess.getinfo("redirector").terminate(timeout=2)

    @pytest.mark.slow
    def test_http_01_server_down(
        self,
        default_order: OrderResource,
        default_account: AccountResource,
        message_builder: MessageBuilder,
    ):
        authz_resource = default_order.authorizations[0]
        chall = authz_resource.challenges[0]

        chall_request = message_builder(acct_id=default_account.id, url=chall.url)

        response = handle(
            ChallengeRequestHandler(
                chall.url, "POST", msg=chall_request, headers=default_request_headers()
            )
        )
        resource = Box(response.msg.model_dump())
        assert resource.status == "processing"
        assert resource.type == "http-01"

        chall_service = inject.instance(IChallengeService)
        chall_service.validate(
            default_account.jwk, default_order, authz_resource, chall
        )

        assert chall.status != ChallengeStatus.valid

    @pytest.mark.slow
    def test_http_01_chall(
        self,
        default_order: OrderResource,
        default_account: AccountResource,
        message_builder: MessageBuilder,
        chall_server_root,
    ):
        authz_resource = default_order.authorizations[0]
        chall = authz_resource.challenges[0]

        chall_request = message_builder(acct_id=default_account.id, url=chall.url)

        response = handle(
            ChallengeRequestHandler(
                chall.url, "POST", msg=chall_request, headers=default_request_headers()
            )
        )
        resource = Box(response.msg.model_dump())
        assert resource.status == "processing"
        assert resource.type == "http-01"

        chall_service = inject.instance(IChallengeService)
        chall_service.validate(
            default_account.jwk, default_order, authz_resource, chall
        )

        assert chall.status != ChallengeStatus.valid

        acme_chall_dir = os.path.join(chall_server_root, ".well-known/acme-challenge/")
        os.makedirs(acme_chall_dir)
        with open(os.path.join(acme_chall_dir, chall.token), "w") as chall_file:
            chall_file.write(
                f"{chall.token}.{josepy.encode_b64jose(default_account.jwk.thumbprint())}"
            )

        chall_service.validate(
            default_account.jwk, default_order, authz_resource, chall
        )

        assert chall.status == ChallengeStatus.valid
        assert chall.validated is not None
        assert authz_resource.status == AuthorizationStatus.valid

    @pytest.mark.slow
    def test_http_01_chall_invalid_cert(
        self,
        default_order: OrderResource,
        default_account: AccountResource,
        message_builder: MessageBuilder,
        https_chall_server_root,
        http_to_https_redirector,
    ):
        authz_resource = default_order.authorizations[0]
        chall = authz_resource.challenges[0]

        chall_request = message_builder(acct_id=default_account.id, url=chall.url)

        response = handle(
            ChallengeRequestHandler(
                chall.url, "POST", msg=chall_request, headers=default_request_headers()
            )
        )
        resource = Box(response.msg.model_dump())
        assert resource.status == "processing"
        assert resource.type == "http-01"

        ValidatorType = validator_for(chall.type)
        validator = ValidatorType(default_account.jwk, authz_resource, chall)

        acme_chall_dir = os.path.join(
            https_chall_server_root, ".well-known/acme-challenge/"
        )
        os.makedirs(acme_chall_dir)
        with open(os.path.join(acme_chall_dir, chall.token), "w") as chall_file:
            chall_file.write(
                f"{chall.token}.{josepy.encode_b64jose(default_account.jwk.thumbprint())}"
            )

        assert validator.validate()



from collections import defaultdict
import random
import string
from typing import Mapping
import uuid

from cryptography.x509 import Certificate, CertificateSigningRequest
from cryptography import x509
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import hashes
import inject
from josepy.jwk import JWK

from acmev2.errors import ACMEError
from acmev2.models import (
    AccountResource,
    OrderResource,
    AuthorizationResource,
    HTTPChallengeResource,
    CustomChallengeResource,
    ChallengeResource,
    ChallengeStatus,
    OrderStatus,
    CertResource,
)
from acmev2.services import (
    IDirectoryService,
    INonceService,
    IAccountService,
    IOrderService,
    IAuthorizationService,
    IChallengeService,
)
from acmev2.services import ACMEEndpoint, ACMEResourceType
from datetime import datetime, timedelta, timezone

from acmev2.services.services import ICertService


class MemoryDirectoryService(IDirectoryService):
    root_url = "http://acme.localhost/acme"
    external_account_required = False

    def url_base(self) -> str:
        return self.root_url


class MemoryNonceService(INonceService):

    nonces = []

    def __init__(self):
        self.nonces = []

    def generate(self) -> str:
        nonce = uuid.uuid4().hex
        self.nonces.append(nonce)

        return nonce

    def consume(self, nonce: str) -> bool:
        if nonce in self.nonces:
            self.nonces.remove(nonce)
            return True

        return False


AccountIdStr = str
KIDStr = str
HMACStr = str


class MemoryAccountService(IAccountService):

    accounts: Mapping[AccountIdStr, tuple[AccountResource, JWK]] = {}
    hmacs: Mapping[KIDStr, HMACStr] = {}
    bound_accounts: Mapping[KIDStr, AccountIdStr] = {}
    order_service = inject.attr(IOrderService)
    account_service = inject.attr(IAccountService)
    directory_service = inject.attr(IDirectoryService)
    authz_service = inject.attr(IAuthorizationService)

    def __init__(self):
        self.accounts = {}
        self.hmacs = {}
        self.bound_accounts = {}

    def gen_hmac(self) -> tuple[KIDStr, HMACStr]:
        kid = "".join(random.choices(string.ascii_letters + string.digits, k=10))
        hmac = "".join(random.choices(string.ascii_letters + string.digits, k=10))

        self.hmacs[kid] = hmac

        return (kid, hmac)

    def create(self, resource: AccountResource, eab_kid: str = None) -> AccountResource:
        if eab_kid and eab_kid in self.bound_accounts:
            raise ACMEError(
                type="malformed",
                detail="Client may not reuse external account bindings",
            )

        while account_id := "".join(
            random.choices(string.ascii_letters + string.digits, k=10)
        ):
            if account_id not in self.accounts:
                break

        resource.id = account_id
        self.accounts[resource.id] = (resource, resource.jwk)

        if eab_kid:
            self.bound_accounts[eab_kid] = resource.id

        return resource

    def get(self, account_id: str) -> AccountResource | None:
        try:
            return self.accounts.get(account_id)[0]
        except:
            return None

    def get_by_jwk(self, jwk: JWK) -> AccountResource | None:
        for acct, acct_jwk in self.accounts.values():
            if jwk.thumbprint() == acct_jwk.thumbprint():
                return acct

    def get_eab_hmac(self, kid: str) -> str | None:
        return self.hmacs.get(kid)

    def check_access(
        self, account_id: str, resource_id: str, resource_type: ACMEResourceType
    ) -> bool:
        # Perms will be done out of band, so we're going to bypass services and
        # just cast to what they are
        order_service: MemoryOrderService = self.order_service
        account_service: MemoryAccountService = self.account_service

        match resource_type:
            case ACMEResourceType.authz:
                for order in order_service.account_orders[account_id]:
                    for auth in order_service.orders[order.id].authorizations:
                        if auth.id == resource_id:
                            return True

                return False
            case ACMEResourceType.challenge:
                for order in order_service.account_orders[account_id]:
                    for auth in order_service.orders[order.id].authorizations:
                        if not auth:
                            continue
                        for chall in auth.challenges:
                            if chall.id == resource_id:
                                return True  # whew

                return False
            case ACMEResourceType.order | ACMEResourceType.cert:
                for order in order_service.account_orders.get(account_id, []):
                    if order.id == resource_id:
                        return True

                return False

            case _:
                return False


OrderIdStr = str


class MemoryOrderService(IOrderService):

    account_orders: Mapping[AccountIdStr, list[OrderResource]] = defaultdict(list)
    orders: Mapping[OrderIdStr, OrderResource] = {}
    order_csrs: Mapping[OrderIdStr, CertificateSigningRequest] = defaultdict(dict)
    authorization_service = inject.attr(IAuthorizationService)

    def __init__(self):
        self.account_orders = defaultdict(list)
        self.orders = {}
        self.order_csrs = defaultdict(dict)

    def create(self, resource: OrderResource) -> OrderResource:
        if not resource.id:
            while order_id := "".join(
                random.choices(string.ascii_letters + string.digits, k=10)
            ):
                if order_id not in self.orders:
                    break

            resource.id = order_id

        self.orders[resource.id] = resource
        self.account_orders[resource.account_id].append(resource)

        return resource

    def update_status(self, order, new_state):
        order.status = new_state
        return order

    def get(self, order_id: str) -> OrderResource | None:
        return self.orders.get(order_id)

    def process_finalization(
        self, order: OrderResource, csr: CertificateSigningRequest
    ) -> OrderResource:
        order.status = OrderStatus.processing
        self.order_csrs[order.id] = csr
        return order


AuthzIdStr = str


class MemoryAuthorizationService(IAuthorizationService):

    authorizations: Mapping[AuthzIdStr, AuthorizationResource] = {}
    challenge_service = inject.attr(IChallengeService)
    directory_service = inject.attr(IDirectoryService)
    order_service = inject.attr(IOrderService)
    authz_service = inject.attr(IAuthorizationService)

    def __init__(self):
        self.authorizations = {}

    def create(self, authz: AuthorizationResource) -> AuthorizationResource:
        if authz.id is None:
            while authz_id := "".join(
                random.choices(string.ascii_letters + string.digits, k=10)
            ):
                if authz_id not in self.authorizations:
                    break

            authz.id = authz_id

        self.authorizations[authz.id] = authz

        return authz

    def update_status(self, authz, new_state):
        authz.status = new_state
        return authz

    def get(self, authz_id: str) -> AuthorizationResource | None:
        return self.authorizations.get(authz_id)

    def get_by_order(self, order_id: str) -> list[AuthorizationResource]:
        order_service: MemoryOrderService = self.order_service
        authorizations: list[AuthorizationResource] = []
        for auth_url in order_service.orders[order_id].authorizations:
            authz_id = self.directory_service.identifier_from_url(
                ACMEEndpoint.authz, auth_url
            )
            authorizations.append(self.authz_service.get(authz_id))

        return authorizations

    def get_by_chall(self, chall_id: str) -> AuthorizationResource | None:
        for _, auth in self.authorizations.items():
            for chall in auth.challenges:
                if chall.id == chall_id:
                    return auth


ChallengeIdStr = str


class MemoryChallengeService(IChallengeService):

    challenges: Mapping[
        ChallengeIdStr, HTTPChallengeResource | CustomChallengeResource
    ] = {}
    directory_service = inject.attr(IDirectoryService)
    authorization_service = inject.attr(IAuthorizationService)

    def __init__(self):
        self.challenges = {}

    def create(self, chall: ChallengeResource) -> list[ChallengeResource]:
        if not chall.id:
            while chall_id := "".join(
                random.choices(string.ascii_letters + string.digits, k=10)
            ):
                if chall_id not in self.challenges:
                    break

            chall.id = chall_id

        self.challenges[chall.id] = chall

        return chall

    def update_status(self, chall, new_state):
        if chall.status != new_state and new_state == ChallengeStatus.valid:
            chall.validated = datetime.now()
        chall.status = new_state

        return chall

    def get(self, chall_id: str) -> ChallengeResource:
        return self.challenges.get(chall_id)

    def queue_validation(self, chall: ChallengeResource) -> ChallengeResource:
        return chall


CA_KEY = rsa.generate_private_key(
    public_exponent=65537,
    key_size=2048,
)

CA_SUBJ = CA_ISSUER = x509.Name(
    [
        x509.NameAttribute(x509.NameOID.COMMON_NAME, "acmeca.local"),
    ]
)
CA_CERT = (
    x509.CertificateBuilder()
    .subject_name(CA_SUBJ)
    .issuer_name(CA_ISSUER)
    .public_key(CA_KEY.public_key())
    .serial_number(x509.random_serial_number())
    .not_valid_before(datetime.now(timezone.utc))
    .not_valid_after(
        # Our certificate will be valid for 10 days
        datetime.now(timezone.utc)
        + timedelta(days=10)
    )
    .add_extension(
        x509.SubjectAlternativeName([x509.DNSName("localhost")]),
        critical=False,
        # Sign our certificate with our private key
    )
    .sign(CA_KEY, hashes.SHA256())
)


class MemoryCertService(ICertService):
    order_service: MemoryOrderService = inject.attr(IOrderService)

    def get(self, ord_id: str) -> CertResource | None:
        csr = self.order_service.order_csrs.get(ord_id)
        if csr:
            builder = (
                x509.CertificateBuilder()
                .subject_name(csr.subject)
                .issuer_name(CA_CERT.issuer)
                .public_key(csr.public_key())
                .serial_number(x509.random_serial_number())
                .not_valid_before(datetime.now(timezone.utc))
                .not_valid_after(datetime.now(timezone.utc) + timedelta(days=90))
            )

            for ext in csr.extensions:
                builder = builder.add_extension(ext.value, ext.critical)

            cert = builder.sign(CA_KEY, hashes.SHA256())

            return CertResource(
                pem=cert.public_bytes(serialization.Encoding.PEM)
                + CA_CERT.public_bytes(serialization.Encoding.PEM)
            )



from datetime import datetime

import inject

from acmev2.errors import (
    ACMEError,
)
from acmev2.messages import ChallengeMessage
from acmev2.models import AuthorizationResource, ChallengeStatus
from acmev2.services import (
    ACMEEndpoint,
    IAuthorizationService,
    IChallengeService,
    IDirectoryService,
)

from .base import ACMEModelResponse, ACMERequestHandler


class ChallengeRequestHandler(ACMERequestHandler):
    """Begins validating a specific challenge. Validation can be done in-line or
    the service implementer can decide to periodically validate in the background."""

    message_type = ChallengeMessage
    directory_service = inject.attr(IDirectoryService)
    challenge_service = inject.attr(IChallengeService)
    authorization_service = inject.attr(IAuthorizationService)

    def process(self, msg: ChallengeMessage):
        chall = msg.resource
        authz = self.authorization_service.get(chall.authz_id)
        self.validate_expiration(msg, authz)

        if chall.status == ChallengeStatus.pending:
            chall = self.challenge_service.queue_validation(chall)
            chall.status = ChallengeStatus.processing

        resp = ACMEModelResponse(msg=chall, code=200)
        resp.add_link_header(
            self.directory_service.url_for(ACMEEndpoint.authz, authz.id), "up"
        )

        return resp

    def validate_expiration(self, msg: ChallengeMessage, authz: AuthorizationResource):
        if datetime.now() > authz.expires:
            raise ACMEError(detail="Authorization has expired")



#!/usr/bin/env python
from datetime import datetime, timedelta, timezone
from http.server import (
    BaseHTTPRequestHandler,
    SimpleHTTPRequestHandler,
    ThreadingHTTPServer,
)

import os
import socket
import ssl
import sys
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import serialization, hashes
from cryptography import x509
from acmev2.tests.helpers import gen_csr

pkey_path: str = None
cert_path: str = None


def _get_best_family(*address):
    infos = socket.getaddrinfo(
        *address,
        type=socket.SOCK_STREAM,
        flags=socket.AI_PASSIVE,
    )
    family, type, proto, canonname, sockaddr = next(iter(infos))
    return family, sockaddr


def test(
    HandlerClass=BaseHTTPRequestHandler,
    ServerClass=ThreadingHTTPServer,
    protocol="HTTP/1.0",
    port=443,
    bind=None,
):

    ServerClass.address_family, addr = _get_best_family(bind, port)
    HandlerClass.protocol_version = protocol
    with ServerClass(addr, HandlerClass) as httpd:
        host, port = httpd.socket.getsockname()[:2]
        url_host = f"[{host}]" if ":" in host else host
        print(
            f"Serving HTTPS on {host} port {port} " f"(https://{url_host}:{port}/) ..."
        )
        try:
            context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
            context.verify_mode = ssl.VerifyMode.CERT_NONE
            context.load_cert_chain(cert_path, pkey_path)
            httpd.socket = context.wrap_socket(
                httpd.socket,
                server_side=True,
            )
            httpd.serve_forever()
        except KeyboardInterrupt:
            print("\nKeyboard interrupt received, exiting.")
            sys.exit(0)


if __name__ == "__main__":
    import argparse
    import contextlib

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-d",
        "--directory",
        default=os.getcwd(),
        help="serve this directory " "(default: current directory)",
    )

    args = parser.parse_args()

    pkey_path = os.path.join(args.directory, "key.pem")
    cert_path = os.path.join(args.directory, "crt.pem")
    key = rsa.generate_private_key(public_exponent=65537, key_size=2048)
    csr = gen_csr(key, "localhost", ["localhost"])
    builder = (
        x509.CertificateBuilder()
        .issuer_name(
            x509.Name(
                [
                    x509.NameAttribute(x509.NameOID.COMMON_NAME, "ca.localhost"),
                ]
            )
        )
        .subject_name(csr.subject)
        .public_key(csr.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.now(timezone.utc))
        .not_valid_after(datetime.now(timezone.utc) + timedelta(days=90))
    )
    cert = builder.sign(key, hashes.SHA256())
    with open(pkey_path, "w") as f:
        f.write(
            key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=serialization.NoEncryption(),
            ).decode()
        )
    with open(cert_path, "w") as f:
        f.write(cert.public_bytes(encoding=serialization.Encoding.PEM).decode())

    handler_class = SimpleHTTPRequestHandler

    # ensure dual-stack is not disabled; ref #38907
    class DualStackServer(ThreadingHTTPServer):

        def server_bind(self):
            # suppress exception when protocol is IPv4
            # with contextlib.suppress(Exception):
            #    self.socket.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 0)
            return super().server_bind()

        def finish_request(self, request, client_address):
            self.RequestHandlerClass(
                request, client_address, self, directory=args.directory
            )

    test(HandlerClass=handler_class, ServerClass=DualStackServer, bind="0.0.0.0")



from box import Box
import inject

from acmev2.models import (
    OrderResource,
    EmptyMessageSchema,
    AuthorizationSchema,
    AuthorizationStatus,
)
from acmev2.services import IAuthorizationService
from acmev2.tests.conftest import GenAccount, AuthorizationRequest
from acmev2.tests.helpers import authz_by_url, resp_has_error, validate_authz
from acmev2.tests.test_base import TestServicesMixin
from acmev2.tests.test_services import MemoryAuthorizationService


class TestAuthorizationRequest(TestServicesMixin):

    def test_post_as_get_authorization(
        self,
        authorization_request: AuthorizationRequest,
        default_order: OrderResource,
    ):
        response = authorization_request(EmptyMessageSchema())
        resource = Box(response.msg.model_dump())

        assert response.code == 200
        assert resource.status == "pending"
        assert resource.expires is not None

        assert resource.identifier.type == default_order.identifiers[0].type
        assert resource.identifier.value == default_order.identifiers[0].value
        assert len(resource.challenges) == 1
        assert resource.challenges[0].status == "pending"
        assert len(resource.challenges[0].token) > 0

    def test_not_found(
        self,
        authorization_request: AuthorizationRequest,
    ):
        response = authorization_request(EmptyMessageSchema(), url="no-resource-found")
        assert response.code == 404

    def test_unauthorized(
        self, authorization_request: AuthorizationRequest, gen_account: GenAccount
    ):
        second_account, second_account_jwk = gen_account()

        response = authorization_request(
            EmptyMessageSchema(), acct_id=second_account.id, jwk_rsa=second_account_jwk
        )

        assert response.code == 401

    def test_deactivate(
        self, default_order: OrderResource, authorization_request: AuthorizationRequest
    ):

        authz = default_order.authorizations[0]
        validate_authz(authz)

        response = authorization_request(
            AuthorizationSchema(status=AuthorizationStatus.deactivated)
        )
        resource = Box(response.msg.model_dump())
        assert resource.status == "deactivated"

    def test_bad_deactivation(self, authorization_request: AuthorizationRequest):
        response = authorization_request(
            AuthorizationSchema(status=AuthorizationStatus.deactivated)
        )
        print(response)
        assert resp_has_error(response, "urn:ietf:params:acme:error:malformed")



from enum import Enum
from typing import Optional

from pydantic import BaseModel, Field

from acmev2.models.challenge import CustomChallengeResource, HTTPChallengeResource

from . import Identifier
from .base import RFC3339Date


class AuthorizationStatus(str, Enum):
    """https://datatracker.ietf.org/doc/html/rfc8555/#section-7.1.6"""

    pending = "pending"
    valid = "valid"
    invalid = "invalid"
    deactivated = "deactivated"
    expired = "expired"
    revoked = "revoked"


class AuthorizationSchema(BaseModel):
    status: Optional[AuthorizationStatus] = None


class AuthorizationResource(BaseModel):
    id: str | None = Field(exclude=True, default=None)
    order_id: str | None = Field(exclude=True, default=None)

    status: AuthorizationStatus = AuthorizationStatus.pending
    expires: RFC3339Date
    identifier: Identifier
    challenges: list[HTTPChallengeResource | CustomChallengeResource] = []



from datetime import datetime, timedelta

import inject
import pytest

from acmev2.handlers import ACMEModelResponse
from acmev2.models import (
    AccountResource,
    AccountStatus,
    Identifier,
    NewOrderRequestSchema,
    OrderResource,
    OrderStatus,
    AuthorizationStatus,
)
from acmev2.services import ACMEEndpoint, IDirectoryService, IOrderService
from acmev2.tests.conftest import NewOrderRequest
from acmev2.tests.helpers import resp_has_error


class TestInvalidAccount:
    @pytest.fixture
    def order_generator(self, new_order_request: NewOrderRequest):
        directory_service = inject.instance(IDirectoryService)
        order_service = inject.instance(IOrderService)

        def wrapped(domains: list[str]):
            response = new_order_request(
                NewOrderRequestSchema(
                    identifiers=[Identifier(type="dns", value=d) for d in domains]
                )
            )
            finalize = response.msg.model_dump()["finalize"]
            ord_id = directory_service.identifier_from_url(
                ACMEEndpoint.finalize, finalize
            )
            return order_service.get(ord_id)

        return wrapped

    def test_expired_order(self, order_generator):
        order_service = inject.instance(IOrderService)
        ord: OrderResource = order_generator(["acme.localhost"])
        ord.expires = datetime.now() - timedelta(minutes=5)
        ord = order_service.resolve_state(ord)

        assert ord.status == OrderStatus.invalid

    def test_expired_authz(self, order_generator):
        order_service = inject.instance(IOrderService)
        ord: OrderResource = order_generator(["acme.localhost"])
        ord.authorizations[0].expires = datetime.now() - timedelta(minutes=5)
        ord = order_service.resolve_state(ord)

        assert ord.status == OrderStatus.invalid
        assert ord.authorizations[0].status == AuthorizationStatus.expired

    def test_do_not_transition_valid_order(self, order_generator):
        order_service = inject.instance(IOrderService)
        ord: OrderResource = order_generator(["acme.localhost"])
        ord.status = OrderStatus.valid
        ord.authorizations[0].expires = datetime.now() - timedelta(minutes=5)
        ord = order_service.resolve_state(ord)

        assert ord.status == OrderStatus.valid

    def test_order_ready(self, order_generator):
        order_service = inject.instance(IOrderService)
        ord: OrderResource = order_generator(["acme.localhost"])

        ord.authorizations[0].status = AuthorizationStatus.valid
        ord = order_service.resolve_state(ord)

        assert ord.status == OrderStatus.ready

    def test_do_not_transition_terminal_authz(self, order_generator):
        order_service = inject.instance(IOrderService)
        ord: OrderResource = order_generator(["acme.localhost"])
        ord.authorizations[0].status = AuthorizationStatus.deactivated
        ord = order_service.resolve_state(ord)

        assert ord.status == OrderStatus.invalid
        assert ord.authorizations[0].status == AuthorizationStatus.deactivated



from enum import Enum
from typing import Optional

import inject
from pydantic import BaseModel, ConfigDict, Field, computed_field

from .base import JoseJsonSchema
import josepy.jwk


class NewAccountRequestSchema(BaseModel):
    termsOfServiceAgreed: Optional[bool] = False
    contact: Optional[list[str]] = []
    onlyReturnExisting: Optional[bool] = False
    externalAccountBinding: Optional[JoseJsonSchema] = None


class AccountStatus(str, Enum):
    valid = "valid"
    deactivated = "deactivated"
    revoked = "revoked"


class AccountResource(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    id: str | None = Field(exclude=True, default=None)

    status: AccountStatus = AccountStatus.valid
    contact: Optional[list[str]] = []
    termsOfServiceAgreed: bool
    contact: Optional[list[str]] = []
    jwk: josepy.jwk.JWK | None = Field(exclude=True, default=None)

    @computed_field
    def orders(self) -> str:
        # avoids circular import...
        from acmev2.services import ACMEEndpoint, IDirectoryService

        return inject.instance(IDirectoryService).url_for(ACMEEndpoint.order, self.id)



import time
import inject
from acmev2.handlers import handle, ChallengeRequestHandler
from acmev2.challenges import validator_for
from acmev2.models import (
    OrderResource,
    AccountResource,
    ChallengeStatus,
    AuthorizationStatus,
)
from acmev2.tests.conftest import MessageBuilder
from acmev2.tests.helpers import authz_by_url
from acmev2.services import IChallengeService, IAccountService
from box import Box
import pytest
from xprocess import ProcessStarter
import tempfile
import os
import asyncio
from acmev2.tests.test_services import MemoryAccountService
import shutil


class TestCaddy:

    CADDYFILE = """
acme.localhost 
  tls {
    ca http://localhost:8080/directory
    ca_root /tmp/acmev2.flask.crt
    eab kid hmac
  }
}
"""

    @pytest.fixture
    def acme_server(self, xprocess):
        app_server = os.path.join(
            os.path.dirname(__file__), "../xprocess/acme_server.py"
        )

        class Starter(ProcessStarter):
            terminate_on_interrupt = True
            timeout = 5
            # startup pattern
            pattern = "Debugger PIN"

            # command to start process
            args = ["uv", "run", app_server]

        # ensure process is running
        xprocess.ensure("acme_server", Starter)

        yield

        # clean up whole process tree afterwards
        xprocess.getinfo("acme_server").terminate()

    @pytest.mark.slow
    def test_caddy(
        self,
        acme_server,
    ):
        # This works intermittently, but I can't stop Caddy from resolving acme.localhost to ipv6 sometimes through
        # the local DNS resolver, even when it's placed in /etc/hosts

        import subprocess

        with open("/tmp/Caddyfile", "w") as cfg:

            cfg.write(self.CADDYFILE)

        cmd = [
            "caddy",
            "run",
        ]

        proc = subprocess.Popen(cmd, cwd="/tmp")
        cert_loc = os.path.expanduser(
            "~/.local/share/caddy/certificates/localhost-8080-directory/acme.localhost/acme.localhost.crt"
        )
        shutil.rmtree(os.path.expanduser("~/.local/share/caddy/"), ignore_errors=True)

        deadline = time.time() + 5
        cert_found = False
        while time.time() < deadline and not cert_found:
            cert_found = os.path.exists(cert_loc)
            time.sleep(1)

        proc.terminate()
        assert cert_found



#!/usr/bin/env python

import contextlib
from http.server import (
    BaseHTTPRequestHandler,
    ThreadingHTTPServer,
)
import socket


class Redirector(BaseHTTPRequestHandler):

    def do_GET(self):
        self.send_response(302)
        new_path = "https://%s%s" % (self.headers["host"], self.path)
        self.send_header("Location", new_path)
        self.end_headers()


# ensure dual-stack is not disabled; ref #38907
class DualStackServer(ThreadingHTTPServer):

    def server_bind(self):
        # suppress exception when protocol is IPv4
        with contextlib.suppress(Exception):
            self.socket.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 0)
        return super().server_bind()


httpd = DualStackServer(("0.0.0.0", 80), Redirector)
print("Serving HTTP->HTTPS redirector")
try:
    httpd.serve_forever()
except KeyboardInterrupt:
    print("\nKeyboard interrupt received, exiting.")
    exit(0)



import inject
import pytest
from acmev2.challenges import CustomChallengeValidator
from acmev2.challenges.base import ChallengeValidator
from acmev2.handlers import ACMEModelResponse
from acmev2.models import (
    NewOrderRequestSchema,
    Identifier,
    ChallengeType,
    AccountResource,
    OrderResource,
    AuthorizationStatus,
    OrderStatus,
    EmptyMessageSchema,
)
from acmev2.settings import ACMESettings, Challenges
from acmev2.tests.conftest import NewOrderRequest, OrderRequest
from acmev2.tests.helpers import authz_by_url
from acmev2.services import (
    IChallengeService,
    IAuthorizationService,
    IOrderService,
    IDirectoryService,
    ACMEEndpoint,
)
from box import Box


class TestAuthStateChange:

    @pytest.fixture()
    def passing_custom_validator(self, setup):
        class AlwaysTrueCustomChallengeValidator(CustomChallengeValidator):
            def validate(self, challenge_validator: ChallengeValidator) -> bool:
                return True

        inject.get_injector()._bindings[
            CustomChallengeValidator
        ] = lambda: AlwaysTrueCustomChallengeValidator()

    @pytest.fixture()
    def failing_custom_validator(self, setup):
        class AlwaysFalseCustomChallengeValidator(CustomChallengeValidator):
            def validate(self, challenge_validator: ChallengeValidator) -> bool:
                return False

        inject.get_injector()._bindings[
            CustomChallengeValidator
        ] = lambda: AlwaysFalseCustomChallengeValidator()

    def test_chall_makes_auth_valid(
        self,
        default_account: AccountResource,
        new_order_request: NewOrderRequest,
        order_request: OrderRequest,
        passing_custom_validator,
    ):
        settings = inject.instance(ACMESettings)
        directory_service = inject.instance(IDirectoryService)
        chall_service = inject.instance(IChallengeService)
        authz_service = inject.instance(IAuthorizationService)
        order_service = inject.instance(IOrderService)
        settings.challenges_available = [Challenges.custom, Challenges.http_01]
        domains = ["domain1.acme.locahost", "domain2.acme.localhost"]
        response: ACMEModelResponse = new_order_request(
            NewOrderRequestSchema(
                identifiers=[Identifier(type="dns", value=domain) for domain in domains]
            )
        )

        resource = Box(response.msg.model_dump())
        assert len(resource.identifiers) == 2
        ord_id = directory_service.identifier_from_url(
            ACMEEndpoint.finalize, resource.finalize
        )

        for authz_url in resource.authorizations:
            authz = authz_by_url(authz_url)
            assert len(authz.challenges) == len(settings.challenges_available)
            for chall in authz.challenges:
                if chall.type == ChallengeType.custom:
                    chall_service.validate(
                        default_account.jwk, order_service.get(ord_id), authz, chall
                    )

            assert authz_service.get(authz.id).status == AuthorizationStatus.valid

        order = order_service.get(ord_id)
        assert order.status == OrderStatus.ready

    @pytest.mark.slow
    def test_chall_makes_auth_invalid(
        self,
        default_account: AccountResource,
        new_order_request: NewOrderRequest,
        order_request: OrderRequest,
        failing_custom_validator,
    ):
        settings = inject.instance(ACMESettings)
        directory_service = inject.instance(IDirectoryService)
        chall_service = inject.instance(IChallengeService)
        authz_service = inject.instance(IAuthorizationService)
        order_service = inject.instance(IOrderService)
        settings.challenges_available = [Challenges.custom, Challenges.http_01]
        domains = ["domain1.acme.locahost", "domain2.acme.localhost"]
        response: ACMEModelResponse = new_order_request(
            NewOrderRequestSchema(
                identifiers=[Identifier(type="dns", value=domain) for domain in domains]
            )
        )

        resource = Box(response.msg.model_dump())
        assert len(resource.identifiers) == 2
        ord_id = directory_service.identifier_from_url(
            ACMEEndpoint.finalize, resource.finalize
        )

        for authz_url in resource.authorizations:
            authz = authz_by_url(authz_url)
            assert len(authz.challenges) == len(settings.challenges_available)
            for chall in authz.challenges:
                if chall.type == ChallengeType.custom:
                    chall_service.validate(
                        default_account.jwk, order_service.get(ord_id), authz, chall
                    )

            assert authz_service.get(authz.id).status == AuthorizationStatus.pending

        order = order_service.get(ord_id)
        assert order.status != OrderStatus.ready



import inject
import requests
import urllib3
from urllib3.exceptions import InsecureRequestWarning

from acmev2.settings import ACMESettings
from .base import ChallengeValidator, ValidationError
import logging
import josepy

logger = logging.getLogger(__name__)


class Http01ChallengeValidator(ChallengeValidator):

    def validate(self) -> bool:
        settings = inject.instance(ACMESettings)
        # Only DNS identifiers supported
        domain = f"http://{self.authz.identifier.value}"

        endpoint = f"{domain}/.well-known/acme-challenge/{self.chall.token}"
        logger.info(f"Validating endpoint <{endpoint}> for challenge <{self.chall.id}>")
        expected_content = (
            f"{self.chall.token}.{josepy.encode_b64jose(self.acct_jwk.thumbprint())}"
        )

        try:
            # warnings can sometimes be treated as errors, we need to allow the challenge
            # to run against certificates that are expired
            urllib3.disable_warnings(InsecureRequestWarning)
            challenge_headers = requests.utils.default_headers()
            challenge_headers["User-Agent"] = settings.http_01_challenge_user_agent
            # again, pass verify=False because this challenge should succeed on invalid certs
            # if the challenge is redirected to https
            challenge_response = requests.get(
                endpoint, allow_redirects=True, verify=False
            )

            if challenge_response.status_code != 200:
                raise ValidationError(
                    "{challenge_response.status_code} status code returned from server"
                )

            logger.info(
                f"Challenge returned: <{challenge_response}>, expected: <{expected_content}>"
            )
        except requests.RequestException as exc:
            logger.exception("Error validating http-01 challenge")
            raise ValidationError(
                f"{exc.__class__.__name__}: error requesting {exc.request.url}"
            )

        # remove extra text returned
        actual = challenge_response.text.splitlines()[0].strip()
        if actual != expected_content:
            raise ValidationError("Content did not match expected token")

        return actual == expected_content



from typing import Type
from .http import Http01ChallengeValidator
from .base import ChallengeValidator
from .custom import CustomChallengeValidatorRunner, CustomChallengeValidator
from acmev2.models import ChallengeType


def validator_for(typ: ChallengeType) -> Type[ChallengeValidator]:
    match typ:
        case ChallengeType.http_01:
            return Http01ChallengeValidator
        case ChallengeType.custom:
            return CustomChallengeValidatorRunner

    raise Exception("No match found for challenge type {typ}")



import inject
from acmev2.errors import ACMEError
from acmev2.messages import AuthorizationMessage
from acmev2.models import AuthorizationStatus
from acmev2.services import IAuthorizationService
from .base import ACMERequestHandler, ACMEModelResponse


class AuthorizationRequestHandler(ACMERequestHandler):
    """Returns an authorization and all the associated challenges."""

    message_type = AuthorizationMessage
    authorization_service = inject.attr(IAuthorizationService)

    def process(self, msg: AuthorizationMessage):
        authz_resource = msg.resource

        if self.should_deactivate(msg):
            authz_resource = self.authorization_service.update_status(
                authz_resource, AuthorizationStatus.deactivated
            )

        return ACMEModelResponse(msg=authz_resource, code=200)

    def should_deactivate(self, msg: AuthorizationMessage) -> bool:
        if msg.payload.status == AuthorizationStatus.deactivated:
            if msg.resource.status == AuthorizationStatus.valid:
                return True

            # Invalid, cannot transition to deactivated from any state but valid
            raise ACMEError(
                detail=f"May not transition to '{AuthorizationStatus.deactivated.value}' from '{msg.resource.status.value}'"
            )

        return False



import pytest
from acmev2.errors import ACMEMethodNotAllowedError
from acmev2.handlers import NewNonceRequestHandler
from acmev2.handlers.handler import handle
from acmev2.services import ACMEEndpoint
from acmev2.tests.helpers import resp_has_error
from acmev2.tests.test_base import TestServicesMixin
from acmev2.tests.conftest import default_request_headers
from acmev2.models import AccountResource, NewOrderRequestSchema, Identifier
from ..conftest import NewOrderRequest


class TestNonceRequest(TestServicesMixin):
    """
    https://datatracker.ietf.org/doc/html/rfc8555/#section-7.2
    Tests getting a new nonce.
    """

    def test_head(self):
        request = NewNonceRequestHandler(
            self.directory_service.url_for(ACMEEndpoint.newNonce),
            "HEAD",
            headers=default_request_headers(),
        )
        response = handle(request)

        nonce = response.headers.get("Replay-Nonce")
        assert nonce is not None
        assert len(nonce) > 0

        assert response.headers.get("Cache-Control") == "no-store"
        assert response.code == 200
        assert response.msg is None

        assert "Link" in response.headers

    def test_get(self):
        request = NewNonceRequestHandler(
            self.directory_service.url_for(ACMEEndpoint.newNonce),
            "GET",
            headers=default_request_headers(),
        )
        response = handle(request)

        nonce = response.headers.get("Replay-Nonce")
        assert nonce is not None
        assert len(nonce) > 0

        assert response.headers.get("Cache-Control") == "no-store"
        assert response.code == 204
        assert response.msg is None

    def test_invalid_verb(self):
        request = NewNonceRequestHandler(
            self.directory_service.url_for(ACMEEndpoint.newNonce),
            "POST",
            headers=default_request_headers(),
        )
        response = handle(request)
        assert resp_has_error(response, "urn:ietf:params:acme:error:malformed")

    def test_bad_nonce(self, new_order_request: NewOrderRequest):
        response = new_order_request(
            NewOrderRequestSchema(
                identifiers=[Identifier(type="dns", value="acme.localhost")]
            ),
            message_builder_kwargs={"nonce": "invalid"},
        )

        assert resp_has_error(response, "urn:ietf:params:acme:error:badNonce")
        assert response.code == 400



from acmev2.models import AuthorizationResource, AuthorizationSchema
from acmev2.services import ACMEResourceType, ACMEEndpoint
from .base import SignedACMEMessage, WithResourceUrl


class AuthorizationMessage(
    SignedACMEMessage[AuthorizationSchema],
    WithResourceUrl[AuthorizationResource],
):
    """https://datatracker.ietf.org/doc/html/rfc8555/#section-7.5"""

    protected_resource_type = ACMEResourceType.authz
    protected_url_type = ACMEEndpoint.authz



from acmev2.models import AccountResource, AccountStatus
from acmev2.handlers import ACMEModelResponse
from acmev2.models import Identifier, NewOrderRequestSchema

from acmev2.tests.conftest import NewOrderRequest
from acmev2.tests.helpers import resp_has_error


class TestInvalidContentType:
    def test_invalid_content_type(self, new_order_request: NewOrderRequest):
        resp = new_order_request(
            NewOrderRequestSchema(
                identifiers=[Identifier(type="dns", value="acme.localhost")]
            )
        )

        assert resp.code == 201

        resp: ACMEModelResponse = new_order_request(
            NewOrderRequestSchema(
                identifiers=[Identifier(type="dns", value="acme.localhost")]
            ),
            request_url="https://invalid.localhost/acme/newOrder",
            headers={"Content-Type": "wrong-type"},
        )

        assert resp.code == 415
        assert resp_has_error(resp, "urn:ietf:params:acme:error:malformed")



import inject
import josepy.jwk

from acmev2.handlers import NewAccountRequestHandler, handle
from acmev2.models.account import AccountResource, NewAccountRequestSchema
from acmev2.services import IAccountService
from acmev2.settings import ACMESettings
from acmev2.tests.conftest import EABMessageBuilder, NewAccountRequest, MessageBuilder
from acmev2.tests.test_services import MemoryAccountService

from ..helpers import (
    gen_jwk_rsa,
    gen_private_key,
    resp_has_error,
)


class TestNewAccountRequest:

    def test_success(self, new_account_request: NewAccountRequest):
        account_service: MemoryAccountService = inject.instance(IAccountService)
        accounts = len(account_service.accounts)

        response = new_account_request(
            NewAccountRequestSchema(termsOfServiceAgreed=True)
        )
        assert len(account_service.accounts) == accounts + 1
        assert response.code == 201
        assert "Location" in response.headers

    def test_force_tos(self, new_account_request: NewAccountRequest):
        # Messages to the new account endpoint fail without
        # agreeing to the ToS
        response = new_account_request(NewAccountRequestSchema())
        assert resp_has_error(response, "urn:ietf:params:acme:error:userActionRequired")

    def test_invalid_contacts(self, new_account_request: NewAccountRequest):
        # Only support mailto: contacts per rfc6068. We don't get too carried
        # away trying to implement the whole ridiculous spec.
        response = new_account_request(
            NewAccountRequestSchema(
                termsOfServiceAgreed=True, contact=["mailto: invalid"]
            )
        )
        assert resp_has_error(response, "urn:ietf:params:acme:error:unsupportedContact")

    def test_only_return_existing_error(self, new_account_request: NewAccountRequest):
        # Only return existing errors if an account does not exist with the passed in jwk
        response = new_account_request(
            NewAccountRequestSchema(termsOfServiceAgreed=True, onlyReturnExisting=True)
        )
        assert resp_has_error(
            response, "urn:ietf:params:acme:error:accountDoesNotExist"
        )

    def test_lookup(self, new_account_request: NewAccountRequest):
        account_service: MemoryAccountService = inject.instance(IAccountService)
        accounts = len(account_service.accounts)

        new_account_request(NewAccountRequestSchema(termsOfServiceAgreed=True))
        lookup_response = new_account_request(
            NewAccountRequestSchema(termsOfServiceAgreed=True)
        )

        assert len(account_service.accounts) == accounts + 1
        assert lookup_response.code == 200
        assert "Location" in lookup_response.headers

    def test_eab_success(
        self,
        jwk_rsa: josepy.JWKRSA,
        new_account_request: NewAccountRequest,
        eab_message_builder: EABMessageBuilder,
    ):
        account_service: MemoryAccountService = inject.instance(IAccountService)
        kid, hmac_key = account_service.gen_hmac()

        response = new_account_request(
            NewAccountRequestSchema(
                termsOfServiceAgreed=True,
                externalAccountBinding=eab_message_builder(kid, hmac_key, jwk_rsa),
            )
        )

        assert response.code == 201
        msg: AccountResource = response.msg
        assert msg.id == account_service.bound_accounts[kid]

    def test_eab_already_bound(
        self,
        new_account_request: NewAccountRequest,
        eab_message_builder: EABMessageBuilder,
    ):
        account_service: MemoryAccountService = inject.instance(IAccountService)
        kid, hmac_key = account_service.gen_hmac()

        def make_request():
            jwk_rsa = gen_jwk_rsa(gen_private_key())

            return new_account_request(
                NewAccountRequestSchema(
                    termsOfServiceAgreed=True,
                    externalAccountBinding=eab_message_builder(kid, hmac_key, jwk_rsa),
                )
            )

        make_request()
        response = make_request()
        assert resp_has_error(response, "urn:ietf:params:acme:error:malformed")

    def test_eab_required(
        self,
        new_account_request: NewAccountRequest,
    ):
        settings = inject.instance(ACMESettings)

        settings.eab_required = True
        response = new_account_request(
            NewAccountRequestSchema(termsOfServiceAgreed=True)
        )

        assert resp_has_error(
            response, "urn:ietf:params:acme:error:externalAccountRequired"
        )

    def test_invalid_eab_kid(
        self,
        jwk_rsa: josepy.JWKRSA,
        new_account_request: NewAccountRequest,
        eab_message_builder: EABMessageBuilder,
    ):
        account_service: MemoryAccountService = inject.instance(IAccountService)
        _, hmac_key = account_service.gen_hmac()

        response = new_account_request(
            NewAccountRequestSchema(
                termsOfServiceAgreed=True,
                externalAccountBinding=eab_message_builder(
                    "invalid-kid", hmac_key, jwk_rsa
                ),
            )
        )

        assert resp_has_error(response, "urn:ietf:params:acme:error:malformed")

    def test_invalid_eab_hmac(
        self,
        jwk_rsa: josepy.JWKRSA,
        new_account_request: NewAccountRequest,
        eab_message_builder: EABMessageBuilder,
    ):
        account_service: MemoryAccountService = inject.instance(IAccountService)
        kid, _ = account_service.gen_hmac()

        response = new_account_request(
            NewAccountRequestSchema(
                termsOfServiceAgreed=True,
                externalAccountBinding=eab_message_builder(
                    kid, "invalid-hmac", jwk_rsa
                ),
            )
        )

        assert resp_has_error(response, "urn:ietf:params:acme:error:malformed")



import inject

from acmev2.errors import ACMEMethodNotAllowedError
from acmev2.messages.base import ACMEMessage
from acmev2.services import INonceService
from .base import ACMERequestHandler, ACMEModelResponse, HttpVerb


class NewNonceRequestHandler(ACMERequestHandler):
    """Returns a new nonce that a client can use on a subsequent request."""

    allowed_verbs = [HttpVerb.GET, HttpVerb.HEAD]
    requires_nonce = False

    @inject.autoparams()
    def process(
        self, msg: ACMEMessage, nonce_service: INonceService = None
    ) -> ACMEModelResponse:
        nonce = nonce_service.generate()

        headers = {"Replay-Nonce": nonce, "Cache-Control": "no-store"}

        if self.verb == HttpVerb.GET:
            code = 204
        elif self.verb == HttpVerb.HEAD:
            code = 200
        else:
            raise ACMEMethodNotAllowedError()

        resp = ACMEModelResponse(headers, code=code)
        return resp



import inject
from acmev2.models import EmptyMessageSchema
from acmev2.services import IDirectoryService
from acmev2.services.services import ACMEEndpoint
from acmev2.tests.conftest import OrderRequest
from acmev2.tests.helpers import resp_has_error


class TestOrder:

    def test_order_post_as_get_contains_location(self, order_request: OrderRequest):
        # All order post-as-get needs to have the 'Location' header. Why? Who knows.
        response = order_request(EmptyMessageSchema())
        assert "Location" in response.headers
        assert response.msg.id is not None

    def test_invalid_account(self, order_request: OrderRequest, gen_account):
        acct, jwk = gen_account()
        response = order_request(
            EmptyMessageSchema(),
            acct_id=acct.id,
            message_builder_kwargs={"jwk_rsa": jwk},
        )

        assert resp_has_error(response, "urn:ietf:params:acme:error:unauthorized")

    def test_account_not_found(self, order_request: OrderRequest):
        url = inject.instance(IDirectoryService).url_for(
            ACMEEndpoint.order, "invalid-id"
        )
        response = order_request(EmptyMessageSchema(), url=url)

        assert response.code == 404
        assert resp_has_error(response, "resourceNotFound")



from enum import Enum

import inject
from pydantic import BaseModel, Field, computed_field, field_serializer

from . import Identifier, AuthorizationResource
from .base import ACMEResource, RFC3339Date


class NewOrderRequestSchema(BaseModel):
    identifiers: list[Identifier]
    # notbefore and notafter are ignored


class OrderFinalizationRequestSchema(BaseModel):
    csr: str


class OrderStatus(str, Enum):
    pending = "pending"
    ready = "ready"
    processing = "processing"
    valid = "valid"
    invalid = "invalid"


class WithFinalizeLink:
    @computed_field
    @property
    def finalize(self: "OrderResource") -> str:
        # avoids circular import...
        from acmev2.services import ACMEEndpoint, IDirectoryService

        return inject.instance(IDirectoryService).url_for(
            ACMEEndpoint.finalize, self.id
        )


class WithCertificateLink:
    @computed_field
    @property
    def certificate(self: "OrderResource") -> str:
        """
        This field is technically optional, but some clients require it to exist. If the cert
        does not exist those clients will try to contact a non-existant url to get the cert
        and will error at retrieval.
        """
        # avoids circular import...
        from acmev2.services import ACMEEndpoint, IDirectoryService

        return inject.instance(IDirectoryService).url_for(ACMEEndpoint.cert, self.id)


class OrderResource(ACMEResource, WithFinalizeLink, WithCertificateLink):
    id: str | None = Field(exclude=True, default=None)
    account_id: str | None = Field(exclude=True, default=None)

    status: OrderStatus = OrderStatus.pending
    expires: RFC3339Date
    identifiers: list[Identifier]
    authorizations: list[AuthorizationResource]
    _mask_processing_status: bool = False

    @field_serializer("status")
    def serialize_status(self, status: OrderStatus, _info):
        # If the client does not support the "processing" state we can trick it
        # into thinking the order is still pending. This is a very specific compensation
        # to fix misbehaving clients, i.e. https://github.com/cert-manager/cert-manager/issues/5062
        if self._mask_processing_status and status == OrderStatus.processing:
            return OrderStatus.pending

        return status

    @field_serializer("authorizations")
    def serialize_authorizations(
        self, authorizations: list[AuthorizationResource], _info
    ):
        from acmev2.services import IDirectoryService, ACMEEndpoint

        directory_service = inject.instance(IDirectoryService)

        return [
            directory_service.url_for(ACMEEndpoint.authz, a.id) for a in authorizations
        ]



import inject
import josepy.b64
import pytest
from acmev2.errors import ACMEBadCSRDetail
from acmev2.models import AccountResource, OrderResource, OrderStatus
from acmev2.models.authorization import AuthorizationStatus
from acmev2.handlers import (
    NewAccountRequestHandler,
    NewOrderRequestHandler,
    CertRequestHandler,
    handle,
    OrderFinalizationRequestHandler,
)
from cryptography.hazmat.primitives import serialization
from acmev2.handlers.base import ACMEModelResponse
from acmev2.services import IDirectoryService, ACMEEndpoint
from acmev2.tests.conftest import MessageBuilder, default_request_headers
from acmev2.tests.helpers import (
    authz_by_url,
    chall_by_url,
    gen_csr,
    gen_jwk_rsa,
    gen_private_key,
    resp_has_error,
    make_order_ready,
)
from box import Box
import josepy.jwk
import josepy
from datetime import datetime, timedelta
from cryptography import x509


class TestCertRequest:

    def test_cert_ready(
        self,
        private_key: josepy.jwk.JWK,
        default_order: OrderResource,
        default_account: AccountResource,
        message_builder: MessageBuilder,
    ):
        directory_service = inject.instance(IDirectoryService)
        make_order_ready(default_order)
        order_request = message_builder(
            {
                "csr": josepy.encode_b64jose(
                    gen_csr(
                        private_key,
                        default_order.identifiers[0].value,
                        [default_order.identifiers[0].value]
                        + [i.value for i in default_order.identifiers],
                    ).public_bytes(serialization.Encoding.DER)
                )
            },
            acct_id=default_account.id,
            url=directory_service.url_for(ACMEEndpoint.finalize, default_order.id),
        )

        handle(
            OrderFinalizationRequestHandler(
                directory_service.url_for(ACMEEndpoint.finalize, default_order.id),
                "POST",
                msg=order_request,
                headers=default_request_headers(),
            )
        )
        cert = handle(
            CertRequestHandler(
                directory_service.url_for(ACMEEndpoint.cert, default_order.id),
                "POST",
                msg=message_builder(
                    acct_id=default_account.id,
                    url=directory_service.url_for(ACMEEndpoint.cert, default_order.id),
                ),
                headers=default_request_headers(),
            )
        )
        assert cert.code == 200
        certs = x509.load_pem_x509_certificates(cert.serialize().encode())
        assert len(certs) == 2
        assert (
            certs[0].subject.rfc4514_string()
            == f"CN={default_order.identifiers[0].value}"
        )

    def test_cert_not_found(
        self,
        default_order: OrderResource,
        default_account: AccountResource,
        message_builder: MessageBuilder,
    ):
        directory_service = inject.instance(IDirectoryService)
        cert = handle(
            CertRequestHandler(
                directory_service.url_for(ACMEEndpoint.cert, default_order.id),
                "POST",
                msg=message_builder(
                    acct_id=default_account.id,
                    url=directory_service.url_for(ACMEEndpoint.cert, default_order.id),
                ),
                headers=default_request_headers(),
            )
        )
        assert cert.code == 404



import abc
import inject
import requests
import urllib3
from urllib3.exceptions import InsecureRequestWarning

from acmev2.settings import ACMESettings
from .base import ChallengeValidator
import logging

logger = logging.getLogger(__name__)


class CustomChallengeValidatorRunner(ChallengeValidator):

    def validate(self) -> bool:
        validator = inject.instance(CustomChallengeValidator)
        return validator.validate(self)


class CustomChallengeValidator(abc.ABC):

    @abc.abstractmethod
    def validate(self, challenge_validator: ChallengeValidator) -> bool:
        pass
